# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Tests for the CUDAQ program type.

"""

import base64
import pathlib
import re

import pytest

try:
    from qbraid import load_program, transpile
    from qbraid.programs.exceptions import ProgramTypeError
    from qbraid.programs.gate_model.cudaq import CudaQKernel
    from qbraid.programs.registry import unregister_program_type

    cudaq_not_installed = False
except ImportError:
    cudaq_not_installed = True

pytestmark = pytest.mark.skipif(cudaq_not_installed, reason="cudaq not installed")


def test_invalid_program_initialization():
    """Tests that initializing a program with an invalid input raises a ProgramTypeError."""
    try:
        with pytest.raises(ProgramTypeError):
            CudaQKernel(("invalid", "program"))
    finally:
        unregister_program_type("tuple")


def test_cudaq_program_num_qubits():
    """Tests that the number of qubits in a CUDAQ program is correctly determined."""

    qasm = """
    OPENQASM 3;
    include "stdgates.inc";
    qubit[2] q;
    bit[2] b;
    h q[0];
    cx q[0], q[1];
    b[0] = measure q[0];
    b[1] = measure q[1];
    """

    kernel = transpile(qasm, "cudaq")

    program = load_program(kernel)

    assert isinstance(program, CudaQKernel)
    assert program.qubits == [0, 1]
    assert program.num_qubits == 2
    assert program.num_clbits == 0


def test_cudaq_program_serialize():
    """Tests that the serialization of a CUDAQ program is correct."""
    import cudaq  # pylint: disable=import-outside-toplevel

    kernel = cudaq.make_kernel()
    qubits = kernel.qalloc(2)
    kernel.h(qubits[0])
    kernel.cx(qubits[0], qubits[1])
    kernel.mz(qubits)

    program = load_program(kernel)

    run_input = program.serialize()
    encoded_qir = run_input["qir"]
    decoded_qir = base64.b64decode(encoded_qir.encode("utf-8")).decode("utf-8")

    test_dir = pathlib.Path(__file__).parent
    fixture_path = test_dir.parent / "fixtures" / "qir" / "bell.ll"

    with open(fixture_path, "r", encoding="utf-8") as file:
        expected_qir = file.read()

    def _normalize_qir(qir: str) -> str:
        return re.sub(r"derKernel_[A-Z0-9]+\(\)", "derKernel_PLACEHOLDER()", qir)

    normalized_expected_qir = _normalize_qir(expected_qir)
    normalized_decoded_qir = _normalize_qir(decoded_qir)

    assert normalized_expected_qir == normalized_decoded_qir



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for converting pytket circuits to Cirq circuits.

"""
import pytest

try:
    import numpy as np
    from pytket.circuit import Circuit as TKCircuit

    from qbraid.interface import circuits_allclose, random_circuit
    from qbraid.transpiler import ConversionGraph, transpile

    pytket_not_installed = False
except ImportError:
    pytket_not_installed = True

pytestmark = pytest.mark.skipif(pytket_not_installed, reason="pytket not installed")


def test_bell_state_from_qiskit():
    """Tests pytket.circuit.Circuit --> cirq.Circuit
    with a Bell state circuit."""
    pytket_circuit = TKCircuit(2)
    pytket_circuit.H(0)
    pytket_circuit.CX(0, 1)
    cirq_circuit = transpile(pytket_circuit, "cirq")
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
def test_crz_gate_from_pytket(qubits):
    """Test converting controlled Rz gate from pytket to cirq."""
    pytket_circuit = TKCircuit(2)
    pytket_circuit.CRz(np.pi / 4, *qubits)
    cirq_circuit = transpile(pytket_circuit, "cirq")
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
@pytest.mark.parametrize("theta", (0, 2 * np.pi, np.pi / 2, np.pi / 4))
def test_rzz_gate_from_pytket(qubits, theta):
    """Test converting Rzz gate from pytket to cirq."""
    pytket_circuit = TKCircuit(2)
    pytket_circuit.ZZPhase(theta, *qubits)
    cirq_circuit = transpile(pytket_circuit, "cirq")
    assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=True)


def test_100_random_pytket():
    """Test converting 100 random pytket circuits to cirq."""
    graph = ConversionGraph(nodes=["pytket", "cirq", "qasm2"])
    for _ in range(100):
        pytket_circuit = random_circuit("pytket", 4, 1, graph=graph)
        cirq_circuit = transpile(pytket_circuit, "cirq")
        assert circuits_allclose(pytket_circuit, cirq_circuit, strict_gphase=False)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for OpenQASM 3 conversions

"""
import braket.circuits
import numpy as np
import qiskit

from qbraid.interface import circuits_allclose
from qbraid.transpiler.converter import transpile


def test_one_qubit_qiskit_to_braket():
    """Test converting qiskit to braket for one qubit circuit."""
    qiskit_circuit = qiskit.QuantumCircuit(1)
    qiskit_circuit.h(0)
    qiskit_circuit.ry(np.pi / 2, 0)
    qasm3_program = transpile(qiskit_circuit, "qasm3")
    braket_circuit = transpile(qasm3_program, "braket")
    circuits_allclose(qiskit_circuit, braket_circuit, strict_gphase=True)


def test_one_qubit_braket_to_qiskit():
    """Test converting braket to qiskit for one qubit circuit."""
    braket_circuit = braket.circuits.Circuit().h(0).ry(0, np.pi / 2)
    qasm3_program = transpile(braket_circuit, "qasm3")
    qiskit_circuit = transpile(qasm3_program, "qiskit")
    assert circuits_allclose(braket_circuit, qiskit_circuit, strict_gphase=True)


def test_two_qubit_braket_to_qiskit():
    """Test converting braket to qiskit for two qubit circuit."""
    braket_circuit = braket.circuits.Circuit().h(0).cnot(0, 1)
    qasm3_program = transpile(braket_circuit, "qasm3")
    qiskit_circuit = transpile(qasm3_program, "qiskit")
    assert circuits_allclose(braket_circuit, qiskit_circuit, strict_gphase=True)


def test_braket_to_qiskit_stdgates():
    """Test converting braket to qiskit for standard gates."""
    circuit = braket.circuits.Circuit()

    circuit.h([0, 1, 2, 3])
    circuit.x([0, 1])
    circuit.y(2)
    circuit.z(3)
    circuit.s(0)
    circuit.si(1)
    circuit.t(2)
    circuit.ti(3)
    circuit.rx(0, np.pi / 4)
    circuit.ry(1, np.pi / 2)
    circuit.rz(2, 3 * np.pi / 4)
    circuit.phaseshift(3, np.pi / 8)
    circuit.v(0)
    circuit.vi(1)
    circuit.iswap(2, 3)
    circuit.swap(0, 2)
    circuit.swap(1, 3)
    circuit.cnot(0, 1)
    circuit.cphaseshift(2, 3, np.pi / 4)

    cirq_circuit = transpile(circuit, "cirq")
    qasm3_program = transpile(circuit, "qasm3")
    qasm2_program = transpile(cirq_circuit, "qasm2")
    qiskit_circuit_1 = transpile(qasm3_program, "qiskit")
    qiskit_circuit_2 = transpile(qasm2_program, "qiskit")
    assert circuits_allclose(circuit, qiskit_circuit_1, strict_gphase=False)
    assert circuits_allclose(circuit, qiskit_circuit_2, strict_gphase=False)


def test_braket_to_qiskit_vi_sxdg():
    """Test converting braket to qiskit with vi/sxdg gate with
    non-strict global phase comparison."""
    circuit = braket.circuits.Circuit().vi(0)
    qasm3_program = transpile(circuit, "qasm3")
    qiskit_circuit = transpile(qasm3_program, "qiskit")
    assert circuits_allclose(circuit, qiskit_circuit, strict_gphase=False)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for logging configuration.

"""

import logging
from unittest.mock import patch

import pytest

from qbraid._logging import DEFAULT_LOG_LEVEL, parse_log_level


@pytest.mark.parametrize(
    "log_level_input, expected_log_level",
    [
        ("DEBUG", logging.DEBUG),
        ("INFO", logging.INFO),
        ("WARNING", logging.WARNING),
        ("ERROR", logging.ERROR),
        ("CRITICAL", logging.CRITICAL),
        ("debug", logging.DEBUG),
        ("10", logging.DEBUG),
        ("20", logging.INFO),
    ],
)
def test_get_log_level_valid_cases(log_level_input, expected_log_level):
    """Test that valid string or numeric log levels are correctly parsed from the environment."""
    assert parse_log_level(log_level_input) == expected_log_level


@pytest.mark.parametrize(
    "log_level_input, invalid_log_level",
    [
        ("INVALID", "INVALID"),
        ("-1", "-1"),
    ],
)
@patch("logging.warning")
def test_get_log_level_invalid_cases(mock_warning, log_level_input, invalid_log_level):
    """Test that invalid log levels fall back to DEFAULT_LOG_LEVEL and raise a warning."""
    assert parse_log_level(log_level_input) == DEFAULT_LOG_LEVEL
    mock_warning.assert_called_once_with(
        "Invalid log level (str) in LOG_LEVEL: %s. Falling back to WARNING.", invalid_log_level
    )


@patch("logging.warning")
def test_get_log_level_empty_env(mock_warning):
    """Test that when LOG_LEVEL is not set, the function defaults to DEFAULT_LOG_LEVEL."""
    assert parse_log_level(None) == DEFAULT_LOG_LEVEL
    mock_warning.assert_not_called(), "No warning should be called when the env var is empty"


@patch("logging.warning")
def test_get_log_level_invalid_int_warning(mock_warning):
    """Test that an invalid integer log level falls back to DEFAULT_LOG_LEVEL and logs a warning."""
    assert parse_log_level("999") == DEFAULT_LOG_LEVEL
    mock_warning.assert_called_once_with(
        "Invalid log level (int) in LOG_LEVEL: %s. Falling back to WARNING.", 999
    )


def test_get_log_level_valid_int():
    """Test that valid integer log levels are correctly parsed from the environment."""
    assert parse_log_level("20") == logging.INFO


def test_get_log_level_case_insensitive():
    """Test that log levels are case insensitive."""
    assert parse_log_level("debug") == logging.DEBUG



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for remote job submissions to the NEC Vector Annealer using the QbraidProvider.

"""
import pytest
from qbraid_core import QbraidSession

from qbraid import QbraidJob, QbraidProvider
from qbraid._logging import logger
from qbraid.runtime import DeviceStatus
from qbraid.runtime.native.result import NECVectorAnnealerResultData
from qbraid.runtime.schemas import QuboSolveParams


def has_access() -> bool:
    """Check if the user has access to the NEC Vector Annealer."""
    try:
        session = QbraidSession()

        user_data = session.get_user()
        permissions_nodes = user_data["permissionsNodes"]
        return "nec.annealer" in permissions_nodes
    except Exception as err:  # pylint: disable=broad-exception-caught
        logger.error(err)

    return False


@pytest.mark.remote
def test_submit_qubo_job_to_nec_vector_annealer():
    """Test submitting a QUBO job to the NEC Vector Annealer."""
    pyqubo = pytest.importorskip("pyqubo", reason="pyqubo is not installed")

    if not has_access():
        pytest.skip("User does not have access to the NEC Vector Annealer")

    s1, s2, s3, s4 = [pyqubo.Spin(f"s{i}") for i in range(1, 5)]
    H = (4 * s1 + 2 * s2 + 7 * s3 + s4) ** 2
    model = H.compile()
    qubo, offset = model.to_qubo()

    params = QuboSolveParams(offset=offset)

    provider = QbraidProvider()

    device = provider.get_device("nec_vector_annealer")

    if device.status() != DeviceStatus.ONLINE:
        pytest.skip("NEC Vector Annealer is not online")

    job: QbraidJob = device.run(qubo, params=params)

    # pylint: disable=no-member
    job.wait_for_final_state()
    result = job.result()
    # pylint: enable=no-member

    result_data = result.data
    assert isinstance(result_data, NECVectorAnnealerResultData)

    if result.success:
        assert result_data.num_solutions == len(result_data.solutions) == 1

        solution = result_data.solutions[0]
        assert isinstance(solution, dict)
        assert {"spin", "energy"}.issubset(solution.keys())

        spin: dict = solution["spin"]
        assert len(spin) == 4
        assert set(spin.keys()) == {"s1", "s2", "s3", "s4"}
        assert sum(spin.values()) > 0
        assert solution["energy"] == 0

    else:
        assert result.details["statusText"] == "Failed to authenticate with NEC server"



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing fixtures for testing

"""
import importlib.util

import numpy as np
import pytest

from qbraid.programs import load_program


def import_circuit_functions(module_name, bell_func_name, shared15_func_name):
    """Attempt to import bell and shared15 functions from a given module."""
    # Construct the full module path assuming the current module is the root
    package = f".{module_name}.circuits"
    try:
        # Attempt to import the module relatively to the current script
        circuits = importlib.import_module(package, __package__)
    except ImportError:
        return None, None  # The module is not installed

    # Get the bell and shared15 functions
    bell_func = getattr(circuits, bell_func_name, None)
    shared15_func = getattr(circuits, shared15_func_name, None)
    return bell_func, shared15_func


# Define the modules and their respective function names
modules_info = {
    "braket": ("braket_bell", "braket_shared15"),
    "pennylane": ("pennylane_bell", "pennylane_shared15"),
    "cirq": ("cirq_bell", "cirq_shared15"),
    "pyquil": ("pyquil_bell", "pyquil_shared15"),
    "qiskit": ("qiskit_bell", "qiskit_shared15"),
    "pytket": ("pytket_bell", "pytket_shared15"),
    "qasm2": ("qasm2_bell", "qasm2_cirq_shared15"),
    "qasm3": ("qasm3_bell", "qasm3_shared15"),
}

bell_circuit_functions = {}
shared15_circuit_functions = {}

# Attempt to import each module and retrieve the bell and shared15 functions
for module, (bell_function_name, shared15_function_name) in modules_info.items():
    bell_function, shared15_function = import_circuit_functions(
        module, bell_function_name, shared15_function_name
    )
    if bell_function:
        bell_circuit_functions[module] = bell_function
    if shared15_function:
        shared15_circuit_functions[module] = shared15_function


@pytest.fixture
def bell_circuit(request):
    """Fixture for getting bell circuit for a given package."""
    package = request.param
    circuit_func = bell_circuit_functions[package]
    circuit = circuit_func()
    return circuit, package


@pytest.fixture
def two_bell_circuits(request):
    """Fixture for getting two bell circuits from two different packages."""
    package1, package2 = request.param
    try:
        circuit1 = bell_circuit_functions[package1]()
    except KeyError:
        circuit1 = None
    try:
        circuit2 = bell_circuit_functions[package2]()
    except KeyError:
        circuit2 = None

    return circuit1, circuit2, package1, package2


@pytest.fixture
def bell_unitary():
    """Fixture containing unitary for bell circuit."""
    return np.array(
        [
            [0.70710678 + 0.0j, 0.0 + 0.0j, 0.70710678 + 0.0j, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.70710678 + 0.0j, 0.0 + 0.0j, 0.70710678 + 0.0j],
            [0.0 + 0.0j, 0.70710678 + 0.0j, 0.0 + 0.0j, -0.70710678 + 0.0j],
            [0.70710678 + 0.0j, 0.0 + 0.0j, -0.70710678 + 0.0j, 0.0 + 0.0j],
        ]
    )


@pytest.fixture
def shared15_circuit(request):
    """Fixture for getting shared15 circuit and unitary for a specific package."""
    package = request.param
    circuit_func = shared15_circuit_functions[package]
    circuit = circuit_func()
    return circuit, package


@pytest.fixture
def two_shared15_circuits(request):
    """Fixture for getting two shared15 circuits from two different packages."""
    package1, package2 = request.param
    circuit1 = shared15_circuit_functions[package1]()
    circuit2 = shared15_circuit_functions[package2]()
    return circuit1, circuit2, package1, package2


@pytest.fixture
def shared15_unitary():
    """Fixture containing unitary for shared15 circuit."""
    if "cirq" in shared15_circuit_functions:
        circuit = shared15_circuit_functions["cirq"]()
    elif "qiskit" in shared15_circuit_functions:
        circuit = shared15_circuit_functions["qiskit"]()
    elif "braket" in shared15_circuit_functions:
        circuit = shared15_circuit_functions["braket"]()
    elif "pennylane" in shared15_circuit_functions:
        circuit = shared15_circuit_functions["pennylane"]()
    elif "pyquil" in shared15_circuit_functions:
        circuit = shared15_circuit_functions["pyquil"]()
    elif "pytket" in shared15_circuit_functions:
        circuit = shared15_circuit_functions["pytket"]()
    else:
        raise ValueError("No shared15 baseline circuit found")

    program = load_program(circuit)
    return program.unitary()


packages_bell = list(bell_circuit_functions.keys())
packages_shared15 = list(shared15_circuit_functions.keys())



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for loading, encoding, and decoding AHS programs using Amazon Braket.

"""
import json

import numpy as np
import pytest
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.atom_arrangement import AtomArrangement
from braket.ahs.driving_field import DrivingField
from braket.ahs.field import Field
from braket.ahs.hamiltonian import Hamiltonian
from braket.ahs.local_detuning import LocalDetuning
from braket.ahs.pattern import Pattern
from braket.timings.time_series import TimeSeries

from qbraid.programs import ProgramTypeError, load_program
from qbraid.programs.ahs import AHSEncoder
from qbraid.programs.ahs.braket_ahs import BraketAHS, BraketAHSDecoder, BraketAHSEncoder


@pytest.fixture
def register_data() -> dict:
    """Return example register data."""
    return {
        "sites": [
            ["0.00000285", "0.00000688050865276332"],
            ["0.00000688050865276332", "0.00000285"],
            ["0.00000688050865276332", "-0.00000285"],
            ["0.00000285", "-0.00000688050865276332"],
            ["-0.00000285", "-0.00000688050865276332"],
            ["-0.00000688050865276332", "-0.00000285"],
            ["-0.00000688050865276332", "0.00000285"],
            ["-0.00000285", "0.00000688050865276332"],
        ],
        "filling": [1, 1, 1, 1, 1, 1, 1, 1],
    }


@pytest.fixture
def hamiltonian_data():
    """Return example Hamiltonian data."""
    return {
        "drivingFields": [
            {
                "amplitude": {
                    "time_series": {
                        "values": ["0.0", "6300000.0", "6300000.0", "0.0"],
                        "times": ["0.0", "1E-7", "0.0000039", "0.000004"],
                    },
                    "pattern": "uniform",
                },
                "phase": {
                    "time_series": {"values": ["0.0", "0.0"], "times": ["0.0", "0.000004"]},
                    "pattern": "uniform",
                },
                "detuning": {
                    "time_series": {
                        "values": ["-31500000.0", "-31500000.0", "31500000.0", "31500000.0"],
                        "times": ["0.0", "1E-7", "0.0000039", "0.000004"],
                    },
                    "pattern": "uniform",
                },
            }
        ],
        "localDetuning": [],
    }


@pytest.fixture
def ahs_data(register_data, hamiltonian_data):
    """Return example AHS data."""
    return {"register": register_data, "hamiltonian": hamiltonian_data}


@pytest.fixture
def ahs_program() -> AnalogHamiltonianSimulation:
    """Return an example AHS program."""
    a = 5.7e-6  # meters

    register = AtomArrangement()
    register.add(np.array([0.5, 0.5 + 1 / np.sqrt(2)]) * a)
    register.add(np.array([0.5 + 1 / np.sqrt(2), 0.5]) * a)
    register.add(np.array([0.5 + 1 / np.sqrt(2), -0.5]) * a)
    register.add(np.array([0.5, -0.5 - 1 / np.sqrt(2)]) * a)
    register.add(np.array([-0.5, -0.5 - 1 / np.sqrt(2)]) * a)
    register.add(np.array([-0.5 - 1 / np.sqrt(2), -0.5]) * a)
    register.add(np.array([-0.5 - 1 / np.sqrt(2), 0.5]) * a)
    register.add(np.array([-0.5, 0.5 + 1 / np.sqrt(2)]) * a)

    time_max = 4e-6  # seconds
    time_ramp = 1e-7  # seconds
    omega_max = 6300000.0  # rad / sec
    delta_start = -5 * omega_max
    delta_end = 5 * omega_max

    omega = TimeSeries()
    omega.put(0.0, 0.0)
    omega.put(time_ramp, omega_max)
    omega.put(time_max - time_ramp, omega_max)
    omega.put(time_max, 0.0)

    delta = TimeSeries()
    delta.put(0.0, delta_start)
    delta.put(time_ramp, delta_start)
    delta.put(time_max - time_ramp, delta_end)
    delta.put(time_max, delta_end)

    phi = TimeSeries().put(0.0, 0.0).put(time_max, 0.0)

    drive = DrivingField(amplitude=omega, phase=phi, detuning=delta)

    return AnalogHamiltonianSimulation(register=register, hamiltonian=drive)


@pytest.fixture
def ahs_program_local_detuning() -> AnalogHamiltonianSimulation:
    """Return an example AHS program with local detuning."""
    a = 5.5e-6  # meters

    register = AtomArrangement()
    register.add([0, 0])
    register.add([a, 0.0])
    register.add([0.5 * a, np.sqrt(3) / 2 * a])

    omega_max = 2.5e6  # rad / seconds
    t_max = np.pi / (np.sqrt(2) * omega_max)  # seconds

    omega = TimeSeries()
    omega.put(0.0, omega_max)
    omega.put(t_max, omega_max)

    phi = TimeSeries().put(0.0, 0.0).put(t_max, 0.0)  # (time [s], value [rad])
    delta_global = TimeSeries().put(0.0, 0.0).put(t_max, 0.0)  # (time [s], value [rad/s])

    drive = DrivingField(amplitude=omega, phase=phi, detuning=delta_global)

    delta_local = TimeSeries()
    delta_local.put(0.0, -omega_max * 20)  # (time [s], value [rad/s])
    delta_local.put(t_max, -omega_max * 20)

    h = Pattern([0, 0, 0.5])

    local_detuning = LocalDetuning(magnitude=Field(time_series=delta_local, pattern=h))

    H = Hamiltonian()
    H += drive
    H += local_detuning

    return AnalogHamiltonianSimulation(hamiltonian=H, register=register)


@pytest.fixture
def ahs_fixture(request):
    """Return an AHS program fixture."""
    if request.param == "ahs_program":
        return request.getfixturevalue("ahs_program")
    if request.param == "ahs_program_local_detuning":
        return request.getfixturevalue("ahs_program_local_detuning")
    raise ValueError("Unsupported fixture")


@pytest.fixture
def encoder() -> BraketAHSEncoder:
    """Return a BraketAHSEncoder object."""
    return BraketAHSEncoder()


@pytest.fixture
def decoder() -> BraketAHSDecoder:
    """Return a BraketAHSDecoder object."""
    return BraketAHSDecoder()


@pytest.mark.parametrize(
    "ahs_fixture", ["ahs_program", "ahs_program_local_detuning"], indirect=True
)
def test_register_serialization(
    ahs_fixture: AnalogHamiltonianSimulation, encoder: BraketAHSEncoder, decoder: BraketAHSDecoder
):
    """Test the serialization and deserialization of an AtomArrangement object."""
    ahs_register_ir = ahs_fixture._register_to_ir()
    ahs_register_data = encoder.encode_register(ahs_register_ir)
    ahs_register_reconstructed = decoder.decode_register(ahs_register_data)
    assert ahs_register_ir == ahs_register_reconstructed


@pytest.mark.parametrize(
    "ahs_fixture", ["ahs_program", "ahs_program_local_detuning"], indirect=True
)
def test_hamiltonian_serialization(
    ahs_fixture: AnalogHamiltonianSimulation, encoder: BraketAHSEncoder, decoder: BraketAHSDecoder
):
    """Test the serialization and deserialization of an AnalogHamiltonianSimulation object."""
    ahs_hamiltonian_ir = ahs_fixture._hamiltonian_to_ir()
    ahs_hamiltonian_data = encoder.encode_hamiltonian(ahs_hamiltonian_ir)
    ahs_hamiltonian_reconstructed = decoder.decode_hamiltonian(ahs_hamiltonian_data)
    assert ahs_hamiltonian_ir == ahs_hamiltonian_reconstructed


@pytest.mark.parametrize(
    "ahs_fixture", ["ahs_program", "ahs_program_local_detuning"], indirect=True
)
def test_ahs_serialization(
    ahs_fixture: AnalogHamiltonianSimulation, encoder: BraketAHSEncoder, decoder: BraketAHSDecoder
):
    """Test the serialization and deserialization of an AnalogHamiltonianSimulation object."""
    ahs_data = encoder.encode_ahs(ahs_fixture)
    ahs_reconstructed = decoder.decode_ahs(ahs_data)
    assert ahs_fixture.to_ir() == ahs_reconstructed.to_ir()


def test_decode_time_series_raises_for_length_mismatch(decoder: BraketAHSDecoder):
    """Test that decoding a TimeSeries with mismatched lengths raises a ValueError."""
    with pytest.raises(ValueError):
        decoder.decode_time_series([0, 1], [0, 1, 2])


def test_braket_ahs_wrapper_attributes(ahs_program, register_data, hamiltonian_data, ahs_data):
    """Test the attributes of the BraketAHS wrapper."""
    program = BraketAHS(ahs_program)
    assert program.num_atoms == program.num_qubits == 8
    assert program.register == register_data
    assert program.hamiltonian == hamiltonian_data
    assert program.to_dict() == ahs_data


def test_ahs_program_encoding(ahs_program, ahs_data):
    """Test that AHSEncoder correctly encodes AnalogHamiltonianProgram."""
    program = BraketAHS(ahs_program)
    encoded_json = json.dumps(program, cls=AHSEncoder)
    assert json.loads(encoded_json) == ahs_data


def test_non_supported_type_encoding():
    """Test that encoding a non-supported type raises a TypeError."""
    with pytest.raises(TypeError):
        json.dumps({1, 2, 3}, cls=AHSEncoder)


def test_fallback_to_superclass():
    """Test that types not handled by AHSEncoder but supported by JSONEncoder are encoded."""
    data = {"key": [1, 2, 3]}
    encoded_json = json.dumps(data, cls=AHSEncoder)
    assert json.loads(encoded_json) == data


def test_load_braket_ahs_program(ahs_program):
    """Test loading a BraketAHS program."""
    program = load_program(ahs_program)
    assert isinstance(program, BraketAHS)


def test_braket_ahs_wrapper_equality(ahs_program, ahs_program_local_detuning, decoder):
    """Test equality comparison of the BraketAHS wrapper."""
    program_1 = BraketAHS(ahs_program)
    program_2 = BraketAHS(ahs_program_local_detuning)
    assert program_1 == BraketAHS(decoder.decode_ahs(program_1.to_dict()))
    assert program_1 != program_2
    assert program_1 != 42


def test_braket_ahs_wrapper_bad_type():
    """Test that creating a BraketAHS wrapper with a bad type raises a ProgramTypeError."""
    with pytest.raises(ProgramTypeError):
        BraketAHS(())



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining qBraid runtime schemas.

.. currentmodule:: qbraid.runtime.schemas

Models
--------

.. autosummary::
   :toctree: ../stubs/

    QbraidSchemaBase
    QbraidSchemaHeader
    DevicePricing
    DeviceData
    ExperimentMetadata
    AhsExperimentMetadata
    AnnealingExperimentMetadata
    GateModelExperimentMetadata
    QuEraQasmSimulationMetadata
    QbraidQirSimulationMetadata
    TimeStamps
    RuntimeJobModel
    QuboSolveParams

Classes
---------

.. autosummary::
   :toctree: ../stubs/

    Credits
    USD

"""
from .base import USD, Credits, QbraidSchemaBase, QbraidSchemaHeader
from .device import DeviceData, DevicePricing
from .experiment import (
    AhsExperimentMetadata,
    AnnealingExperimentMetadata,
    ExperimentMetadata,
    GateModelExperimentMetadata,
    QbraidQirSimulationMetadata,
    QuboSolveParams,
    QuEraQasmSimulationMetadata,
)
from .job import RuntimeJobModel, TimeStamps

__all__ = [
    "QbraidSchemaBase",
    "QbraidSchemaHeader",
    "DevicePricing",
    "DeviceData",
    "AhsExperimentMetadata",
    "AnnealingExperimentMetadata",
    "ExperimentMetadata",
    "GateModelExperimentMetadata",
    "QbraidQirSimulationMetadata",
    "QuboSolveParams",
    "QuEraQasmSimulationMetadata",
    "RuntimeJobModel",
    "TimeStamps",
    "Credits",
    "USD",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing functions to convert between OpenQASM 3 and IonQ JSON format.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import pyqasm

from qbraid._logging import logger
from qbraid.programs.gate_model.ionq import IONQ_NATIVE_GATES
from qbraid.transpiler.annotations import weight
from qbraid.transpiler.conversions.openqasm3.openqasm3_to_ionq import openqasm3_to_ionq
from qbraid.transpiler.exceptions import ProgramConversionError

if TYPE_CHECKING:
    from qbraid.programs.typer import IonQDictType, Qasm3StringType


@weight(1)
def qasm3_to_ionq(qasm: Qasm3StringType) -> IonQDictType:
    """
    Convert an OpenQASM 3 string to IonQ JSON format.

    Args:
        qasm (Qasm3StringType): The OpenQASM 3 string to convert.

    Returns:
        IonQDictType: IonQ JSON format equivalent to the input OpenQASM 3 string.

    Raises:
        ProgramConversionError: If the conversion fails.
    """
    # pylint: disable=broad-exception-caught
    try:
        return openqasm3_to_ionq(qasm)
    except Exception as err:
        err_msg = str(err) or "Failed to convert OpenQASM 3 to IonQ JSON."
        cache_err = None
        try:
            if not any(gate in qasm for gate in IONQ_NATIVE_GATES):
                module = pyqasm.loads(qasm)
                module.unroll()
                return openqasm3_to_ionq(pyqasm.dumps(module))
        except Exception as pyqasm_err:  # pragma: no cover
            logger.debug("Conversion with pyqasm assistance failed: %s", pyqasm_err)
            cache_err = pyqasm_err
        if cache_err and all(
            msg not in err_msg
            for msg in [
                "Cannot mix native and QIS gates in the same circuit",
                "Circuits with measurements are not supported by the IonQDictType",
            ]
        ):
            err_msg += "." if err_msg and not err_msg.endswith(".") else ""
        else:
            cache_err = err
        raise ProgramConversionError(err_msg) from cache_err
    # pylint: enable=broad-exception-caught



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name,unused-argument

"""
Unit tests for OQCProvider class

"""
from __future__ import annotations

import datetime
import json
import logging
import os
import textwrap
import uuid
from typing import Optional, Union
from unittest.mock import MagicMock, Mock, patch

import pytest
from requests import ReadTimeout

from qbraid.runtime.enums import DeviceStatus
from qbraid.runtime.exceptions import ResourceNotFoundError

try:
    from qcaas_client.client import OQCClient, QPUTask, QPUTaskErrors, QPUTaskResult  # type: ignore

    from qbraid.programs import NATIVE_REGISTRY, ExperimentType, ProgramSpec
    from qbraid.runtime import GateModelResultData, Result, TargetProfile
    from qbraid.runtime.enums import DeviceStatus, JobStatus
    from qbraid.runtime.exceptions import ResourceNotFoundError
    from qbraid.runtime.oqc import OQCDevice, OQCJob, OQCProvider
    from qbraid.runtime.postprocess import counts_to_probabilities
    from qbraid.runtime.schemas.base import USD

    FIXTURE_COUNT = sum(key in NATIVE_REGISTRY for key in ["qiskit", "braket", "cirq"])

    oqc_extras_installed = True
except ImportError as err:
    FIXTURE_COUNT = 0

    oqc_extras_installed = False

    logging.warning("OQC runtime tests will be skipped: %s", err)

pytestmark = pytest.mark.skipif(oqc_extras_installed is False, reason="qcaas_client not installed")

LUCY_SIM_ID = "qpu:uk:2:d865b5a184"
TOSHIKO_ID = "qpu:jp:3:673b1ad43c"

LUCY_FEATURE_SET = {"always_on": True, "qubit_count": 8, "simulator": True}
TOSHIKO_FEATURE_SET = {
    "qubit_count": 32,
    "Maximum Shots": 100000,
    "Maximum Repetition Period": "1 second",
    "Maximum Shots x Repetition Period": "600 seconds",
}

LUCY_SIM_MOCK_DATA = {
    "active": True,
    "feature_set": json.dumps(LUCY_FEATURE_SET),
    "generation": 2,
    "id": LUCY_SIM_ID,
    "name": "Lucy Simulator",
    "price_per_shot": "0",
    "price_per_task": "0",
    "region": "uk",
    "status": "ACTIVE",
    "url": "https://uk.cloud.oqc.app/d865b5a184",
}

TOSHIKO_MOCK_DATA = {
    "active": True,
    "feature_set": json.dumps(TOSHIKO_FEATURE_SET),
    "generation": 3,
    "id": TOSHIKO_ID,
    "name": "OQC Toshiko Tokyo-1",
    "price_per_shot": "0.0013",
    "price_per_task": "0",
    "region": "jp",
    "status": "ACTIVE",
    "url": "https://jp.cloud.oqc.app/673b1ad43c",
}

MOCK_TIMINGS = {
    "RECEIVER_DEQUEUED": "2023-10-17 11:24:32.937188+00:00",
    "RECEIVER_ENQUEUED": "2023-10-17 11:24:32.996594+00:00",
    "RECEIVER_FROM_SCC": "2023-10-17 11:24:32.996594+00:00",
    "RECEIVER_TO_SCC": "2023-10-17 11:24:32.938188+00:00",
    "SERVER_DEQUEUED": "2023-10-17 11:24:36.057355+00:00",
    "SERVER_ENQUEUED": "2023-10-17 11:24:35.946476+00:00",
    "SERVER_RECEIVED": "2023-10-17 11:24:35.904885+00:00",
}

MOCK_METRICS = {"optimized_circuit": "dummy", "optimized_instruction_count": 42}

MOCK_TASK_ID = "e35fb436-ff08-44c8-8acc-7d5a1f1a0ada"
MOCK_TASK_ID_NO_RESULT = "e35fb436-ff08-44c8-8acc-7d5a1f1a0adb"

MOCK_TASK_METADATA = {
    "allow_support_access": False,
    "config": '{"$type": "<class \'scc.compiler.config.CompilerConfig\'>", "$data": {"repeats": null, "repetition_period": null, "results_format": {"$type": "<class \'scc.compiler.config.QuantumResultsFormat\'>", "$data": {"format": null, "transforms": {"$type": "<enum \'scc.compiler.config.ResultsFormatting\'>", "$value": 2}}}, "metrics": {"$type": "<enum \'scc.compiler.config.MetricsType\'>", "$value": 6}, "active_calibrations": [], "optimizations": null, "error_mitigation": null}}',  # pylint: disable=line-too-long
    "created_at": "Wed, 12 Jun 2024 20:11:15 GMT",
    "id": MOCK_TASK_ID,
    "qpu_id": LUCY_SIM_ID,
    "tag": None,
}

MOCK_RESULT = {"c": {"00": 52, "11": 48}}

NOW = datetime.datetime.now()
NEXT_YEAR = NOW.year + 1

TOSHIKO_EXEC_ESTIMATE = {
    "qpu_wait_times": [
        {
            "average_processing_seconds": 0.0,
            "estimated_availability_time": f"{NEXT_YEAR}-11-18 00:50:00",
            "qpu_id": "qpu:jp:3:673b1ad43c",
            "tasks_in_queue": 21,
            "timestamp": f"{NOW.year}-11-13 16:20:43.229464",
            "windows": [
                {
                    "end_time": f"{NEXT_YEAR}-11-18 04:00:00",
                    "start_time": f"{NEXT_YEAR}-11-18 00:50:00",
                    "window_description": "NEXT",
                },
                {
                    "end_time": f"{NEXT_YEAR}-11-18 13:00:00",
                    "start_time": f"{NEXT_YEAR}-11-18 09:50:00",
                    "window_description": "FUTURE",
                },
                {
                    "end_time": f"{NEXT_YEAR}-11-19 04:00:00",
                    "start_time": f"{NEXT_YEAR}-11-19 00:50:00",
                    "window_description": "FUTURE",
                },
                {
                    "end_time": f"{NEXT_YEAR}-11-19 13:00:00",
                    "start_time": f"{NEXT_YEAR}-11-19 09:50:00",
                    "window_description": "FUTURE",
                },
            ],
        }
    ]
}


def online_window() -> str:
    """Return a window start time for an online QPU."""
    now = datetime.datetime.now()
    start_time = f"{now.year}-{now.month}-{now.day} {(now.hour):02}:00:00"
    end_time = f"{now.year}-{now.month}-{now.day} {(now.hour):02}:59:59"
    return start_time, end_time


@pytest.fixture
def program():
    """Return a QASM3 program."""
    qasm3 = """
    OPENQASM 3;
    include "stdgates.inc";
    qubit[2] q;
    bit[2] c;

    h q[0];
    cx q[0], q[1];

    c = measure q;
    """

    qasm3_compat = textwrap.dedent(qasm3).strip()
    return qasm3_compat


@pytest.fixture
def optimized_program():
    """Return a QASM3 program."""
    qasm3 = """
    OPENQASM 3.0;
    qubit[2] q;
    bit[2] c;
    h q[0];
    cx q[0], q[1];
    c = measure q;
    """
    qasm3_compat = textwrap.dedent(qasm3).strip()
    return qasm3_compat


class MockOQCClient:
    """Test class for OQC client."""

    def __init__(self, authentication_token=None, **kwargs):
        self._authentication_token = authentication_token
        self.default_qpu_id = "qpu:uk:3:9829a5504f"
        self.url = "https://cloud.oqc.app/"
        self._toshiko_online = False

    def get_qpus(self):
        """Get QPUs."""
        return [LUCY_SIM_MOCK_DATA.copy(), TOSHIKO_MOCK_DATA.copy()]

    def schedule_tasks(
        self,
        tasks: Union[QPUTask, list[QPUTask]],
        qpu_id: Optional[str] = None,
        tag: Optional[str] = None,
    ):
        """Schedule tasks for the QPU."""
        if not isinstance(tasks, list):
            tasks = [tasks]

        first_qpu_id = None
        assigned_mock_task_id = False

        for task in tasks:
            if not isinstance(task, QPUTask):
                raise ValueError("All tasks must be of type QPUTask")

            if qpu_id:
                task.qpu_id = qpu_id
            else:
                if not task.qpu_id:
                    task.qpu_id = self.default_qpu_id

                if first_qpu_id is None:
                    first_qpu_id = task.qpu_id
                elif task.qpu_id != first_qpu_id:
                    raise ValueError("All tasks must have the same qpu_id")

            if not task.task_id:
                if not assigned_mock_task_id:
                    task.task_id = MOCK_TASK_ID
                    assigned_mock_task_id = True
                else:
                    task.task_id = str(uuid.uuid4())

        return tasks

    def get_next_window(self, qpu_id: Optional[str] = None):
        """Get next window."""
        if qpu_id == LUCY_SIM_ID:
            start_time, _ = online_window()
            return start_time
        raise ReadTimeout

    def get_qpu_execution_estimates(self, qpu_ids: Optional[str] = None):
        """Get QPU execution estimates."""
        if qpu_ids == TOSHIKO_ID:
            exec_est = TOSHIKO_EXEC_ESTIMATE.copy()
            if self._toshiko_online:
                start_time, end_time = online_window()
                current_window = {
                    "end_time": end_time,
                    "start_time": start_time,
                    "window_description": "CURRENT",
                }
                new_windows = [current_window] + exec_est["qpu_wait_times"][0]["windows"]
                exec_est["qpu_wait_times"][0]["windows"] = new_windows
            return exec_est
        raise Exception("QPU execution estimates not available")

    def get_task_status(self, task_id: str, qpu_id: Optional[str] = None):
        """Get task status."""
        if task_id == MOCK_TASK_ID:
            return "COMPLETED"
        return "FAILED"

    def get_task_timings(self, task_id: str, qpu_id: Optional[str] = None):
        """Get task timings."""
        return MOCK_TIMINGS

    def get_task_metrics(self, task_id: str, qpu_id: Optional[str] = None):
        """Get task metrics."""
        return MOCK_METRICS

    def get_task_metadata(self, task_id: str, qpu_id: Optional[str] = None):
        """Get task metadata."""
        metadata = MOCK_TASK_METADATA.copy()
        metadata["id"] = task_id
        if qpu_id:
            metadata["qpu_id"] = qpu_id
        return metadata

    def get_task_errors(self, task_id: str, qpu_id: Optional[str] = None):
        """Get task errors."""
        if task_id == MOCK_TASK_ID:
            return None
        return QPUTaskErrors("Error", 400)

    def get_task_results(self, task_id: str, qpu_id: Optional[str] = None):
        """Get task results."""
        qpu_id_matches = qpu_id is None or qpu_id == LUCY_SIM_ID
        result = MOCK_RESULT if task_id == MOCK_TASK_ID and qpu_id_matches else None
        metrics = self.get_task_metrics(task_id, qpu_id)
        error_details = self.get_task_errors(task_id, qpu_id)
        return QPUTaskResult(task_id, result=result, metrics=metrics, error_details=error_details)

    def cancel_task(self, task_id: str, qpu_id: Optional[str] = None):
        """Cancel task."""
        return None


@pytest.fixture
def lucy_sim_id():
    """Return Lucy Simulator ID."""
    return LUCY_SIM_ID


@pytest.fixture
def toshiko_id():
    """Return Toshiko ID."""
    return TOSHIKO_ID


@pytest.fixture
def lucy_sim_data():
    """Return data for Lucy Simulator."""
    return LUCY_SIM_MOCK_DATA.copy()


@pytest.fixture
def toshiko_data():
    """Return data for Toshiko QPU."""
    return TOSHIKO_MOCK_DATA.copy()


@pytest.fixture
def mock_job_id():
    """Return a mock task ID."""
    return MOCK_TASK_ID


@pytest.fixture
def oqc_client():
    """Return a fake OQC client."""
    return MockOQCClient("fake_token")


@pytest.fixture
def target_profile(lucy_sim_data):
    """Return a target profile for Lucy Simulator."""
    feature_set = LUCY_FEATURE_SET.copy()
    simulator = feature_set.pop("simulator")
    num_qubits = feature_set.pop("qubit_count")

    return TargetProfile(
        device_id=lucy_sim_data["id"],
        device_name=lucy_sim_data["name"],
        simulator=simulator,
        experiment_type=ExperimentType.GATE_MODEL,
        endpoint_url=lucy_sim_data["url"],
        num_qubits=num_qubits,
        program_spec=ProgramSpec(str, alias="qasm3"),
        feature_set=feature_set,
        price_per_shot=USD(lucy_sim_data["price_per_shot"]),
        price_per_task=USD(lucy_sim_data["price_per_task"]),
        generation=lucy_sim_data["generation"],
        region=lucy_sim_data["region"],
    )


@pytest.fixture
def oqc_device(oqc_client, target_profile):
    """Return a fake OQC device."""
    return OQCDevice(profile=target_profile, client=oqc_client)


@pytest.fixture
def oqc_job(mock_job_id, oqc_client, oqc_device):
    """Return a fake OQC job."""
    return OQCJob(mock_job_id, client=oqc_client, device=oqc_device)


@pytest.fixture
def oqc_job_failed(oqc_client, oqc_device):
    """Return a fake OQC job that failed."""
    return OQCJob("failed_job_id", client=oqc_client, device=oqc_device)


def test_oqc_device_str_representation(oqc_device):
    """Test the string representation of an OQC device."""
    assert str(oqc_device) == "OQCDevice('Lucy Simulator')"


def test_oqc_provider_get_devices(lucy_sim_data, toshiko_data):
    """Test getting all OQC devices."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]
        provider = OQCProvider(token="fake_token")
        assert isinstance(provider, OQCProvider)
        assert isinstance(provider.client, OQCClient)
        assert provider.client == mock_client.return_value
        devices = provider.get_devices()
        assert isinstance(devices, list)
        assert {device.id for device in devices} == {lucy_sim_data["id"], toshiko_data["id"]}


def test_oqc_provider_get_device_profile(lucy_sim_data, toshiko_data):
    """Test getting a specific OQC device."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]
        provider = OQCProvider(token="fake_token")
        test_device = provider.get_device(lucy_sim_data["id"])
        assert isinstance(test_device.status(), DeviceStatus)
        assert isinstance(test_device, OQCDevice)
        assert test_device.profile["device_id"] == lucy_sim_data["id"]
        assert test_device.profile["price_per_shot"] == USD(lucy_sim_data["price_per_shot"])
        assert test_device.profile["price_per_task"] == USD(lucy_sim_data["price_per_task"])
        assert isinstance(test_device.client, OQCClient)


def test_oqc_provider_get_device_status_offline(lucy_sim_data, toshiko_data):
    """Test getting a specific OQC device."""
    toshiko_data["active"] = False
    toshiko_data["status"] = "INACTIVE"
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]
        provider = OQCProvider(token="fake_token")
        test_device = provider.get_device(toshiko_data["id"])
        assert test_device.status() == DeviceStatus.OFFLINE


def test_oqc_device_status_from_window_unavailable(lucy_sim_data, toshiko_data):
    """Test that device status value varies correctly based on next available window."""
    lucy_sim_data["feature_set"] = json.dumps(
        {"always_on": False, "qubit_count": 8, "simulator": True}
    )
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]
        provider = OQCProvider(token="fake_token")

        now = datetime.datetime.now()
        year, month, day = now.year, now.month, now.day
        window = f"{year + 1}-{month}-{day} 00:50:00"
        mock_client.return_value.get_next_window.return_value = window
        unavailable_device = provider.get_device(lucy_sim_data["id"])
        assert unavailable_device.status() == DeviceStatus.UNAVAILABLE


def test_oqc_device_status_always_on(lucy_sim_data, toshiko_data):
    """Test that device status is ONLINE for devices that are always on."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]
        provider = OQCProvider(token="fake_token")
        device = provider.get_device(lucy_sim_data["id"])
        assert device.status() == DeviceStatus.ONLINE
        mock_client.get_next_window.assert_not_called()


def test_get_next_window_read_timeout(toshiko_id):
    """Test getting the next window with a timeout results in status unavialable."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = MockOQCClient()
        provider = OQCProvider(token="fake_token")
        device = provider.get_device(toshiko_id)
        assert device.status() == DeviceStatus.UNAVAILABLE


def test_oqc_device_status_from_qpu_exec_est_unavailable(lucy_sim_data, toshiko_data):
    """Test that device status value varies correctly based on next available window."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]

        provider = OQCProvider(token="fake_token")
        qpu_device = provider.get_device(toshiko_data["id"])
        qpu_device._client = MockOQCClient(authentication_token="fake_token")
        assert qpu_device.status() == DeviceStatus.UNAVAILABLE


def test_oqc_device_status_from_qpu_exec_est_online(lucy_sim_data, toshiko_data):
    """Test that device status value varies correctly based on next available window."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]

        provider = OQCProvider(token="fake_token")
        qpu_device = provider.get_device(toshiko_data["id"])
        qpu_device._client = MockOQCClient(authentication_token="fake_token")
        qpu_device._client._toshiko_online = True
        assert qpu_device.status() == DeviceStatus.ONLINE
        assert qpu_device.queue_depth() == 21


def test_oqc_queue_depth_raises_for_simulator(lucy_sim_data, toshiko_data):
    """Test that the queue depth method raises an error for a simulator device."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]

        provider = OQCProvider(token="fake_token")
        simulator_device = provider.get_device(lucy_sim_data["id"])
        with pytest.raises(ResourceNotFoundError) as excinfo:
            simulator_device.queue_depth()
        assert "Queue depth is not available for this device." in str(excinfo.value)


def test_oqc_provider_get_device_raises_not_found(lucy_sim_data, toshiko_data):
    """Test raising an error when a device is not found."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]
        provider = OQCProvider(token="fake_token")
        with pytest.raises(ResourceNotFoundError):
            provider.get_device("fake_id")


def test_oqc_device_status_raises(lucy_sim_data, toshiko_data):
    """Test OQC device status method raises an error for bad request data."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]

        provider = OQCProvider(token="fake_token")
        fake_profile = TargetProfile(
            device_id="fake_id",
            device_name="Fake Device",
            simulator=True,
            experiment_type=ExperimentType.GATE_MODEL,
            endpoint_url="https://uk.cloud.oqc.app/fake_id",
            num_qubits=8,
            program_spec=ProgramSpec(str, alias="qasm3"),
        )
        fake_device = OQCDevice(profile=fake_profile, client=provider.client)
        with pytest.raises(ResourceNotFoundError):
            fake_device.status()


@pytest.mark.parametrize(
    "data_modifications, expected_error_message",
    [
        ({"feature_set": ""}, "Failed to decode feature set data"),
        (
            {"feature_set": json.dumps({"simulator": True})},
            "Failed to gather profile data for device",
        ),
    ],
)
def test_oqc_provider_get_device_errors(lucy_sim_data, data_modifications, expected_error_message):
    """Test OQC provider get device method raises exceptions for various error conditions."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        invalid_lucy_sim_data = lucy_sim_data
        invalid_lucy_sim_data.update(data_modifications)
        mock_client.return_value.get_qpus.return_value = [invalid_lucy_sim_data]
        provider = OQCProvider(token="fake_token")

        with pytest.raises(ValueError) as excinfo:
            provider.get_device(lucy_sim_data["id"])
        assert expected_error_message in str(excinfo.value)


def test_build_runtime_profile(lucy_sim_data):
    """Test building a runtime profile for OQC device."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        provider = OQCProvider(token="fake_token")
        profile = provider._build_profile(lucy_sim_data)
        assert isinstance(profile, TargetProfile)
        assert profile["device_id"] == lucy_sim_data["id"]
        assert profile["simulator"] is True
        assert profile["num_qubits"] == 8
        assert profile["program_spec"] == ProgramSpec(str, alias="qasm3")


@pytest.mark.parametrize("circuit", range(FIXTURE_COUNT), indirect=True)
def test_run_fake_job(circuit, oqc_device):
    """Test running a fake job."""
    job: OQCJob = oqc_device.run(circuit, shots=100)
    assert isinstance(job, OQCJob)
    assert isinstance(job.status(), JobStatus)
    assert isinstance(job.get_timings(), dict)
    assert isinstance(job.metrics(), dict)
    assert isinstance(job.metrics()["optimized_instruction_count"], int)
    assert isinstance(job.metadata(), dict)
    assert isinstance(job.get_errors(), (str, type(None)))

    result = job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, GateModelResultData)
    assert result.data.get_counts() == MOCK_RESULT["c"]


def test_run_batch_fake_job(run_inputs, oqc_device):
    """Test running a batch of fake jobs."""
    job = oqc_device.run(run_inputs)
    assert isinstance(job, list)
    assert len(job) == len(run_inputs)
    assert all(isinstance(j, OQCJob) for j in job)


def test_cancel_job(oqc_device, program):
    """Test cancelling a fake job."""
    job = oqc_device.run(program, shots=100)
    assert isinstance(job, OQCJob)
    assert job.cancel() is None


@pytest.mark.remote
def test_oqc_runtime_remote_execution(program, optimized_program):
    """Test OQC runtime with remote execution."""
    token = os.getenv("OQC_AUTH_TOKEN")
    if token is None:
        pytest.skip("Missing OQC_AUTH_TOKEN")

    provider = OQCProvider(token=token)

    device = provider.get_device(LUCY_SIM_ID)
    assert isinstance(device, OQCDevice)
    assert device.status() == DeviceStatus.ONLINE

    shots = 100
    job = device.run(program, shots=shots)
    assert isinstance(job, OQCJob)

    job.wait_for_final_state()
    assert job.qpu_id == LUCY_SIM_ID
    assert job.is_terminal_state()
    assert job.status() == JobStatus.COMPLETED

    result = job.result()
    assert isinstance(result, Result)
    assert result.details["errors"] is None
    assert result.details["shots"] == shots

    optimized_out = result.details["metrics"]["optimized_circuit"]
    assert optimized_out.strip() == optimized_program.strip()

    data = result.data
    assert isinstance(data, GateModelResultData)

    counts = data.get_counts()
    assert len(counts) == 2
    assert set(counts.keys()) == {"00", "11"}
    assert sum(counts.values()) == shots
    assert data.measurements is None


def test_oqc_job_get_errors(oqc_job_failed):
    """Test getting errors from a failed job."""
    errors = oqc_job_failed.get_errors()
    assert errors == {"message": "Error", "code": 400}

    result = oqc_job_failed.result()
    assert result.success is False
    assert result.data.measurement_counts is None
    assert result.data.measurements is None


@patch("builtins.hash", autospec=True)
def test_hash_method_creates_and_returns_hash(mock_hash, oqc_client):
    """Test that the hash method creates and returns a hash."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        provider_instance = OQCProvider(token="fake_token")
        mock_hash.return_value = 9999
        provider_instance.client = oqc_client
        result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
        mock_hash.assert_called_once_with(("https://cloud.oqc.app/", "fake_token"))
        assert result == 9999
        assert provider_instance._hash == 9999


def test_hash_method_returns_existing_hash(oqc_client):
    """Test that the hash method returns an existing hash."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        provider_instance = OQCProvider(token="fake_token")
        provider_instance.client = oqc_client
        provider_instance._hash = 1234
        result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
        assert result == 1234


def test_job_result_method_raises_resource_not_found(oqc_job):
    """Test that the result method raises a ResourceNotFoundError when no result is found."""
    oqc_job._qpu_id = "mismatched_id"
    with pytest.raises(ResourceNotFoundError) as excinfo:
        oqc_job.result()
    assert "No result found for the task" in str(excinfo.value)


def test_job_get_counts_single_key():
    """Test that the _get_counts method returns the expected counts for a single key."""
    result = {"c": {"00": 1000, "01": 500}}
    expected = {"00": 1000, "01": 500}

    assert OQCJob._get_counts(result) == expected


@pytest.fixture
def multi_cbit_result_counts():
    """Return a result counts dictionary with multiple keys."""
    return {
        "c0": {"000000": 45, "111111": 55},
        "c1": {"000000": 45, "111111": 55},
        "c2": {"000000": 45, "111111": 55},
    }


def test_job_get_counts_multiple_keys(multi_cbit_result_counts):
    """Test that the _get_counts method returns the expected counts for multiple keys."""
    expected = [
        {"000000": 45, "111111": 55},
        {"000000": 45, "111111": 55},
        {"000000": 45, "111111": 55},
    ]

    assert OQCJob._get_counts(multi_cbit_result_counts) == expected


def test_job_get_counts_empty_dict():
    """Test that the _get_counts method raises a ValueError when the result dictionary is empty."""
    result = {}

    with pytest.raises(ValueError, match="The result dictionary must not be empty."):
        OQCJob._get_counts(result)


def test_job_get_counts_list_to_probabilities(multi_cbit_result_counts):
    """Test that the _get_counts method returns the expected probabilities for a list of counts."""
    expected = [
        {"000000": 0.45, "111111": 0.55},
        {"000000": 0.45, "111111": 0.55},
        {"000000": 0.45, "111111": 0.55},
    ]

    counts_list = OQCJob._get_counts(multi_cbit_result_counts)
    assert counts_to_probabilities(counts_list) == expected


def test_job_get_qpu_id_from_task_metadata(lucy_sim_id, oqc_job, oqc_client):
    """Test getting the qpu_id from the task metadata."""
    oqc_job._qpu_id = None
    oqc_job._device = None
    oqc_job._client = oqc_client
    assert oqc_job.qpu_id == lucy_sim_id

    oqc_job._qpu_id = None
    oqc_job._device = None
    oqc_client.get_task_metadata = MagicMock(return_value={"qpu_id": lucy_sim_id})
    oqc_job._client = oqc_client
    assert oqc_job.qpu_id == lucy_sim_id
    oqc_client.get_task_metadata.assert_called_once()


def test_oqc_provider_raises_for_no_token(monkeypatch):
    """Test that the OQCProvider raises an error when no token is provided."""
    monkeypatch.setenv("OQC_AUTH_TOKEN", "")

    with pytest.raises(ValueError) as excinfo:
        OQCProvider()
    assert "An OQC authenication token is required to initialize the provider." in str(
        excinfo.value
    )


@patch("qbraid.runtime.oqc.device.logger")
@patch("qbraid.runtime.oqc.device.OQCDevice.get_next_window")
def test_device_status_online(mock_get_next_window, mock_logger):
    mock_get_next_window.return_value = datetime.datetime(2023, 10, 31, 12, 0, 0)
    device = OQCDevice(profile=Mock(), client=Mock())
    result = device.status()
    assert result == DeviceStatus.ONLINE
    mock_logger.error.assert_not_called()


def test_build_compiler_config_unsupported_key():
    """Test that the build_compiler_config method raises an error for unsupported kwargs."""
    with pytest.raises(ValueError, match="Unsupported keyword argument"):
        OQCDevice._build_compiler_config(unsupported_kwarg="value")


def test_build_compiler_config_invalid_value():
    """Test that the build_compiler_config method raises an error for invalid values."""
    with pytest.raises(ValueError, match="Invalid configuration option"):
        OQCDevice._build_compiler_config(optimizations="invalid_value")


@patch("qbraid.runtime.oqc.device.logger")
def test_device_get_next_window_raises_resource_not_found(mock_logger, target_profile):
    """Test that the get_next_window method raises a ResourceNotFoundError when the window is not found."""
    client = Mock()
    client.get_next_window.side_effect = ReadTimeout
    client.get_qpu_execution_estimates.side_effect = Exception
    device = OQCDevice(target_profile, client)
    with pytest.raises(ResourceNotFoundError) as excinfo:
        device.get_next_window()
    assert "Falied to fetch next active window for device" in str(excinfo.value)
    mock_logger.error.assert_called_once()


@patch("qbraid.runtime.oqc.device.logger")
def test_catch_device_status_resource_not_found(mock_logger, lucy_sim_data, toshiko_data):
    """Test that device status is unavailable when get_next_window method raises error."""
    with patch("qbraid.runtime.oqc.provider.OQCClient") as mock_client:
        mock_client.return_value = Mock(spec=OQCClient)
        mock_client.return_value.get_qpus.return_value = [lucy_sim_data, toshiko_data]

        provider = OQCProvider(token="fake_token")
        device = provider.get_device(toshiko_data["id"])

        with patch.object(device, "get_next_window", side_effect=ResourceNotFoundError):
            status = device.status()
            assert status == DeviceStatus.UNAVAILABLE
            mock_logger.info.assert_called_once()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for submissions to IonQ devices via qBraid native runtime.

"""
import os
import warnings

import pytest

from qbraid import GateModelResultData, QbraidJob, QbraidProvider


@pytest.mark.remote
def test_qiskit_ionq_workflow():
    """Test the workflow of running a Qiskit circuit on an IonQ device via qBraid."""
    try:
        # pylint: disable=import-outside-toplevel
        import qiskit
        import qiskit.qasm2
        import qiskit_ionq

        # pylint: enable=import-outside-toplevel

        current_file_directory = os.path.dirname(os.path.abspath(__file__))
        path_to_qasm_file = os.path.join(current_file_directory, "test.qasm")
        qiskit_circuit = qiskit.qasm2.load(path_to_qasm_file)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            qiskit_ionq_provider = qiskit_ionq.IonQProvider()
            qiskit_ionq_backend = qiskit_ionq_provider.get_backend("ionq_simulator")

            qiskit_circuit_transpiled = qiskit.transpile(qiskit_circuit, qiskit_ionq_backend)

        provider = QbraidProvider()
        device = provider.get_device("ionq_simulator")

        shots = 10
        job: QbraidJob = device.run(qiskit_circuit_transpiled, shots=shots)

        # pylint: disable=no-member
        job.wait_for_final_state()
        result = job.result()
        # pylint: enable=no-member

        assert result.success
        assert isinstance(result.data, GateModelResultData)

        counts: dict[str, int] = result.data.get_counts()
        assert sum(counts.values()) == shots
    except ImportError as err:
        pytest.skip(f"Skipped test due to import error: {err}")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for QASM preprocessing functions

"""
import re
import textwrap

import pytest

from qbraid.passes.qasm.compat import (
    _evaluate_expression,
    _normalize_case_insensitive_map,
    add_stdgates_include,
    convert_qasm_pi_to_decimal,
    has_redundant_parentheses,
    insert_gate_def,
    normalize_qasm_gate_params,
    remove_stdgates_include,
    replace_gate_names,
    simplify_parentheses_in_qasm,
)


@pytest.mark.parametrize(
    "qasm3_without, qasm3_with",
    [
        (
            """
OPENQASM 3;
qubit[1] q;
h q[0];
ry(pi/4) q[0];
        """,
            """
OPENQASM 3;
include "stdgates.inc";
qubit[1] q;
h q[0];
ry(pi/4) q[0];
        """,
        ),
    ],
)
def test_add_stdgates_include(qasm3_without, qasm3_with):
    """Test adding stdgates include to OpenQASM 3.0 string"""

    def _remove_empty_lines(input_string: str) -> str:
        """Removes all empty lines from the provided string."""
        return "\n".join(line for line in input_string.split("\n") if line.strip())

    test_with = _remove_empty_lines(add_stdgates_include(qasm3_without))
    expected_with = _remove_empty_lines(qasm3_with)

    test_without = _remove_empty_lines(remove_stdgates_include(qasm3_with))
    expected_without = _remove_empty_lines(qasm3_without)

    test_redundant = add_stdgates_include(qasm3_with)

    assert test_with == expected_with
    assert test_without == expected_without
    assert test_redundant == qasm3_with


@pytest.mark.parametrize(
    "qasm3_str_pi, qasm3_str_decimal",
    [
        (
            """
        OPENQASM 3;
        qubit[2] q;
        h q[0];
        rx(pi / 4) q[0];
        ry(2*pi) q[0];
        rz(3 * pi/4) q[0];
        cry(pi/4/2) q[0], q[1];
        """,
            """
        OPENQASM 3;
        qubit[2] q;
        h q[0];
        rx(0.7853981633974483) q[0];
        ry(6.283185307179586) q[0];
        rz(2.356194490192345) q[0];
        cry(0.39269908169872414) q[0], q[1];
        """,
        ),
    ],
)
def test_convert_pi_to_decimal(qasm3_str_pi, qasm3_str_decimal):
    """Test converting pi symbol to decimal in qasm3 string"""
    qasm_out = normalize_qasm_gate_params(qasm3_str_pi)
    assert qasm_out == qasm3_str_decimal


@pytest.mark.parametrize(
    "qasm_body, old_gate, new_gate, expected_body",
    [
        (
            "cnot q[0], q[1];",
            "cnot",
            "cx",
            "cx q[0], q[1];",
        ),
        (
            "cnot q[0], q[1];",
            "cnot",
            "custom",
            "custom q[0], q[1];",
        ),
        (
            "cnot q[0], q[1];",
            "notmatched",
            "cx",
            "cnot q[0], q[1];",
        ),
    ],
)
def test_replace_gate_name_qubit_2(qasm_body, old_gate, new_gate, expected_body):
    """Test replacing gate names in QASM strings for two qubit gates."""
    qasm = f"OPENQASM 3; qubit[2] q; {qasm_body}"
    expected_output = f"OPENQASM 3;\nqubit[2] q;\n{expected_body}\n"
    assert replace_gate_names(qasm, {old_gate: new_gate}) == expected_output

    qasm2 = f"OPENQASM 2; qreg q[2]; {qasm_body}"
    qasm2_expected_output = f"OPENQASM 2;\nqreg q[2];\n{expected_body}\n"
    assert replace_gate_names(qasm2, {old_gate: new_gate}) == qasm2_expected_output


@pytest.mark.parametrize(
    "qasm_body, old_gate, new_gate, expected_body",
    [
        (
            "x q[0];",
            "x",
            "pauli_x",
            "pauli_x q[0];",
        ),
        (
            "p(3.14) q[0];",
            "p",
            "phaseshift",
            "phaseshift(3.14) q[0];",
        ),
        (
            "v q[0];",
            "v",
            "sx",
            "sx q[0];",
        ),
        (
            "ti q[0];",
            "ti",
            "tdg",
            "tdg q[0];",
        ),
    ],
)
def test_replace_gate_name_qubit_1(qasm_body, old_gate, new_gate, expected_body):
    """Test replacing gate names in QASM strings for one qubit gates."""
    qasm3 = f"OPENQASM 3; qubit[1] q; {qasm_body}"
    qasm3_expected_output = f"OPENQASM 3;\nqubit[1] q;\n{expected_body}\n"
    assert replace_gate_names(qasm3, {old_gate: new_gate}) == qasm3_expected_output

    qasm2 = f"OPENQASM 2; qreg q[1]; {qasm_body}"
    qasm2_expected_output = f"OPENQASM 2;\nqreg q[1];\n{expected_body}\n"
    assert replace_gate_names(qasm2, {old_gate: new_gate}) == qasm2_expected_output


@pytest.mark.parametrize(
    "qasm3_in, qasm3_out",
    [
        (
            """
        OPENQASM 3.0;
        qubit[2] q;
        cnot q[0], q[1];
        sx q[0];
        p(3.14) q[1];
        cp(3.14) q[0], q[1];
        """,
            """
        OPENQASM 3.0;
        qubit[2] q;
        cx q[0], q[1];
        v q[0];
        phaseshift(3.14) q[1];
        cphaseshift(3.14) q[0], q[1];
        """,
        ),
    ],
)
def test_parameterized_replacement(qasm3_in, qasm3_out):
    """Test replacing non-parameterized gates with parameterized gates in qasm3 string"""
    replacements = {"cnot": "cx", "sx": "v", "p": "phaseshift", "cp": "cphaseshift"}
    qasm3 = replace_gate_names(qasm3_in, replacements)
    qasm3 = textwrap.dedent(qasm3).strip()
    qasm3_out = textwrap.dedent(qasm3_out).strip()
    assert qasm3 == qasm3_out


def test_bad_insert_gate_def():
    """Test inserting gate definition with invalid qasm3 string"""
    qasm3 = ""
    with pytest.raises(ValueError):
        insert_gate_def(qasm3, "bad_gate")


def test_simplify_parentheses_in_qasm():
    """Test simplifying parentheses in qasm string"""
    qasm = """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
cry((0.7853981633974483)) q[0], q[1];
ry(-(0.39269908169872414)) q[1];
"""
    expected = """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
cry(0.7853981633974483) q[0], q[1];
ry(-0.39269908169872414) q[1];
"""
    assert simplify_parentheses_in_qasm(qasm).strip() == expected.strip()


@pytest.mark.parametrize(
    "qasm_input, expected_result",
    [
        (
            """
    OPENQASM 3;
    include "stdgates.inc";
    qubit[2] q;
    cry((0.39269908169872414)) q[0], q[1];
    """,
            True,
        ),
        (
            """
    OPENQASM 3;
    include "stdgates.inc";
    qubit[2] q;
    crx(-(0.39269908169872414)) q[0], q[1];
    """,
            True,
        ),
        (
            """
    OPENQASM 3;
    include "stdgates.inc";
    qubit[2] q;
    h q[0];
    rx(0.7853981633974483) q[0];
    ry(6.283185307179586) q[0];
    rz(2.356194490192345) q[0];
    cry(0.39269908169872414) q[0], q[1];
    """,
            False,
        ),
    ],
)
def test_has_redundant_parentheses(qasm_input, expected_result):
    """Test checking for redundant parentheses in QASM string."""
    assert has_redundant_parentheses(qasm_input) == expected_result


def test_evaluate_expression_error():
    """Test simplifying arithmetic expression with error in qasm string"""
    qasm_str = "gate str(gate_name)(a, b) q {U(1*) q;}"
    match = re.search(r"\(([0-9+\-*/. ]+)\)", qasm_str)

    assert _evaluate_expression(match) == "(1*)"


def test_convert_qasm_pi_to_decimal_omits_gpi_gate():
    """Test converting pi symbol to decimal in qasm string with gpi and gpi2 gates."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    gpi(0) q[0];
    gpi2(0) q[1];
    cry(pi) q[0], q[1];
    """

    expected = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    gpi(0) q[0];
    gpi2(0) q[1];
    cry(3.141592653589793) q[0], q[1];
    """
    assert convert_qasm_pi_to_decimal(qasm) == expected


def test_convert_qasm_pi_to_decimal_gpi2_iso():
    """Test converting pi symbol to decimal in qasm string with gpi2 gate on its own."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[1];
    gpi2(pi/4) q[0];
    """

    expected = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[1];
    gpi2(0.7853981633974483) q[0];
    """
    assert convert_qasm_pi_to_decimal(qasm) == expected


def test_convert_qasm_pi_to_decimal_qasm3_fns_gates_vars():
    """Test converting pi symbol to decimal in a qasm3 string
    with custom functions, gates, and variables."""
    qasm = """
    OPENQASM 3;
    include "stdgates.inc";

    gate pipe q {
        gpi(0) q;
        gpi2(pi/8) q;
    }

    const int[32] primeN = 3;
    const float[32] c = primeN*pi/4;
    qubit[3] q;

    def spiral(qubit[primeN] q_func) {
    for int i in [0:primeN-1] { pipe q_func[i]; }
    }

    spiral(q);

    ry(c) q[0];

    bit[3] result;
    result = measure q;
    """

    expected = """
    OPENQASM 3;
    include "stdgates.inc";

    gate pipe q {
        gpi(0) q;
        gpi2(0.39269908169872414) q;
    }

    const int[32] primeN = 3;
    const float[32] c = primeN*0.7853981633974483;
    qubit[3] q;

    def spiral(qubit[primeN] q_func) {
    for int i in [0:primeN-1] { pipe q_func[i]; }
    }

    spiral(q);

    ry(c) q[0];

    bit[3] result;
    result = measure q;
    """
    assert convert_qasm_pi_to_decimal(qasm) == expected


def test_forced_gate_def_insertion():
    """Test inserting gate definition with force_insert=True."""
    qasm = """
OPENQASM 3.0;
include "stdgates.inc";
qubit[2] q;
h q[0];
cnot q[0], q[1];
    """

    expected = """
OPENQASM 3.0;
gate iswap _gate_q_0, _gate_q_1 {
  s _gate_q_0;
  s _gate_q_1;
  h _gate_q_0;
  cx _gate_q_0, _gate_q_1;
  cx _gate_q_1, _gate_q_0;
  h _gate_q_1;
}
include "stdgates.inc";
qubit[2] q;
h q[0];
cnot q[0], q[1];
    """
    assert insert_gate_def(qasm, "iswap", force_insert=True) == expected


def test_gate_def_insertion_with_include():
    """Test inserting gate definition when include statement is present (include_idx case)."""
    qasm = """
OPENQASM 3.0;
include "stdgates.inc";
qubit[1] q;
h q[0];
sxdg q[0];
    """

    expected = """
OPENQASM 3.0;
gate sxdg _gate_q_0 {
  s _gate_q_0;
  h _gate_q_0;
  s _gate_q_0;
}
include "stdgates.inc";
qubit[1] q;
h q[0];
sxdg q[0];
    """
    result = insert_gate_def(qasm, "sxdg")
    assert result == expected


def test_gate_def_insertion_without_include():
    """Test inserting gate definition when no include statement is present (openqasm_idx case)."""
    qasm = """
OPENQASM 3.0;
qubit[1] q;
h q[0];
    """

    expected = """
OPENQASM 3.0;
gate sxdg _gate_q_0 {
  s _gate_q_0;
  h _gate_q_0;
  s _gate_q_0;
}
qubit[1] q;
h q[0];
    """
    result = insert_gate_def(qasm, "sxdg", force_insert=True)
    assert result == expected


def test_normalize_case_insensitive_map():
    """Test normalizing a case-insensitive map."""
    test_map = {"a": 1, "B": 2, "c": 3}
    expected_map = {"a": 1, "b": 2, "c": 3}
    assert _normalize_case_insensitive_map(test_map) == expected_map


def test_normalize_case_insensitive_map_raises_error():
    """Test normalizing a case-insensitive map raises an error if keys are not unique."""
    test_map = {"a": 1, "A": 2}
    with pytest.raises(ValueError):
        _normalize_case_insensitive_map(test_map)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QiskitJob Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from qbraid_core._import import LazyLoader
from qiskit.primitives.containers.primitive_result import PrimitiveResult
from qiskit_ibm_runtime import RuntimeJobV2
from qiskit_ibm_runtime.exceptions import RuntimeInvalidStateError

from qbraid._logging import logger
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.exceptions import JobStateError, QbraidRuntimeError
from qbraid.runtime.job import QuantumJob
from qbraid.runtime.result import Result
from qbraid.runtime.result_data import GateModelResultData

from .result_builder import QiskitGateModelResultBuilder

if TYPE_CHECKING:
    import qiskit.primitives
    import qiskit.result
    import qiskit_ibm_runtime

    import qbraid.runtime.ibm

qbraid_rt_ibm: qbraid.runtime.ibm = LazyLoader("qbraid_rt_ibm", globals(), "qbraid.runtime.ibm")

IBM_JOB_STATUS_MAP = {
    "INITIALIZING": JobStatus.INITIALIZING,
    "QUEUED": JobStatus.QUEUED,
    "VALIDATING": JobStatus.VALIDATING,
    "RUNNING": JobStatus.RUNNING,
    "CANCELLED": JobStatus.CANCELLED,
    "DONE": JobStatus.COMPLETED,
    "ERROR": JobStatus.FAILED,
}


class QiskitJob(QuantumJob):
    """Class for interfacing with a Qiskit IBM ``RuntimeJob``."""

    def __init__(
        self,
        job_id: str,
        job: Optional[
            qiskit_ibm_runtime.RuntimeJob
            | qiskit_ibm_runtime.RuntimeJobV2
            | qiskit.primitives.PrimitiveJob
        ] = None,
        service: Optional[qiskit_ibm_runtime.QiskitRuntimeService] = None,
        **kwargs,
    ):
        """Create a ``QiskitJob`` instance."""
        super().__init__(job_id, **kwargs)
        self._job = job or self._get_job(service=service)

    def _get_job(
        self, service: Optional[qiskit_ibm_runtime.QiskitRuntimeService] = None
    ) -> qiskit_ibm_runtime.RuntimeJob | qiskit_ibm_runtime.RuntimeJobV2:
        """Return the qiskit_ibm_runtime.RuntimeJob associated with instance id attribute.

        Attempts to retrieve a job using a specified or default service. Handles
        service initialization with error management for authentication issues.

        Raises:
            QbraidRuntimeError: If there is an error initializing the service or retrieving the job.
        """
        if service is None:
            if self._device and getattr(self._device, "_service", None):
                service = self._device._service
            else:
                try:
                    service = qbraid_rt_ibm.QiskitRuntimeProvider().runtime_service
                except Exception as err:
                    raise QbraidRuntimeError("Failed to initialize the quantum service.") from err

        try:
            return service.job(self.id)
        except Exception as err:
            raise QbraidRuntimeError(f"Error retrieving job {self.id}") from err

    def status(self):
        """Returns status from Qiskit Job object."""
        job_status = self._job.status()
        job_status = job_status if isinstance(job_status, str) else job_status.name
        status = IBM_JOB_STATUS_MAP.get(job_status, JobStatus.UNKNOWN)
        self._cache_metadata["status"] = status
        return status

    def queue_position(self) -> Optional[int]:
        """Returns the position of the job in the server queue."""
        if isinstance(self._job, RuntimeJobV2):
            raise NotImplementedError("Queue position is not supported for RuntimeJobV2.")
        return self._job.queue_position(refresh=True)

    def result(self):
        """Return the results of the job."""
        if not self.is_terminal_state():
            logger.info("Result will be available when job has reached final state.")

        runner_result: qiskit.result.Result | PrimitiveResult = self._job.result()
        if isinstance(runner_result, PrimitiveResult):
            backend = self._job.backend() if hasattr(self._job, "backend") else None
            device_id = (
                backend.name if backend else (self._device.id if self._device else "sampler")
            )
            job_id = self._job.job_id()
            success = self.status() == JobStatus.COMPLETED
            runner_data = runner_result.metadata
        else:
            runner_data = runner_result.to_dict()
            job_id = runner_data.pop("job_id", self._job.job_id())
            success = runner_data.pop("success", runner_result.success)
            device_id = runner_result.backend_name

        builder = QiskitGateModelResultBuilder(runner_result)
        measurement_counts = builder.get_counts()
        measurements = builder.measurements()
        data = GateModelResultData(measurement_counts=measurement_counts, measurements=measurements)

        return Result(
            device_id=device_id,
            job_id=job_id,
            success=success,
            data=data,
            **runner_data,
        )

    def cancel(self):
        """Attempt to cancel the job."""
        if self.is_terminal_state():
            raise JobStateError("Cannot cancel quantum job in non-terminal state.")
        try:
            return self._job.cancel()
        except RuntimeInvalidStateError as err:
            raise JobStateError from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unt tests for conversions to/from pyQuil circuits.

"""
import numpy as np
import pytest
from cirq import Circuit, LineQubit
from cirq import ops as cirq_ops
from cirq.contrib.qasm_import import circuit_from_qasm

try:
    from pyquil import Program
    from pyquil.gates import CNOT, CZ, RX, RZ, RZZ, H, I, X, Y, Z
    from pyquil.noise import _decoherence_noise_model, _get_program_gates, apply_noise_model

    from qbraid.interface import circuits_allclose
    from qbraid.transpiler.conversions.cirq import cirq_to_pyquil
    from qbraid.transpiler.conversions.pyquil import pyquil_to_cirq
    from qbraid.transpiler.conversions.qasm2 import qasm2_to_cirq
    from qbraid.transpiler.exceptions import ProgramConversionError

    pyquil_not_installed = False
except ImportError:
    pyquil_not_installed = True

pytestmark = pytest.mark.skipif(pyquil_not_installed, reason="pyquil not installed")


def test_to_from_pyquil():
    """Test round trip pyQuil-Cirq conversions."""
    p = Program()
    p += X(0)
    p += Y(1)
    p += Z(2)
    p += CNOT(0, 1)
    p += CZ(1, 2)
    p_cirq = pyquil_to_cirq(p)
    p_test = cirq_to_pyquil(p_cirq)
    assert p.out() == p_test.out()


def test_to_from_pyquil_parameterized():
    """Test round trip pyQuil-Cirq conversions with parameterized gates."""
    q0, q1 = (0, 1)
    p = Program()
    p += H(q0)
    p += H(q1)
    p += CNOT(q0, q1)
    p += RZ(2 * np.pi, q1)
    p += CNOT(q0, q1)
    p += H(q0)
    p += H(q1)
    p += RZ(np.pi / 4, q0)
    p += RZ(np.pi / 4, q1)
    p += H(q0)
    p += H(q1)
    p_test = cirq_to_pyquil(pyquil_to_cirq(p))
    assert p.out() == p_test.out()


QUIL_STRING = """
I 0
I 1
I 2
X 0
Y 1
Z 2
RX(pi/2) 0
RY(pi/2) 1
RZ(pi/2) 2
H 0
CZ 0 1
CNOT 1 2
CPHASE(pi/2) 0 1
CPHASE00(pi/2) 1 2
CPHASE01(pi/2) 0 1
CPHASE10(pi/2) 1 2
ISWAP 0 1
SWAP 1 2
XY(pi/2) 0 1
CCNOT 0 1 2
CSWAP 0 1 2
"""


def test_to_from_pyquil_quil_string():
    """PHASE, PSWAP, S, T, declaration, and measurement don't convert back
    and forth perfectly (in terms of labels -- the program unitaries and
    number of measurements are equivalent)."""
    p = Program(QUIL_STRING)
    p_cirq = pyquil_to_cirq(p)
    p_cirq = circuit_from_qasm(p_cirq.to_qasm())
    p_test = cirq_to_pyquil(p_cirq)
    assert circuits_allclose(p, p_test)


def test_from_pyquil_no_zero_qubit():
    """Test converting a pyQuil program with a non-zero qubit index to Cirq."""
    p = Program()
    p += X(10)
    p += Y(11)
    p += Z(12)
    p += CNOT(10, 11)
    p += CZ(11, 12)
    p_test = cirq_to_pyquil(pyquil_to_cirq(p))
    assert p.out() == p_test.out()


def test_raise_error_to_pyquil_bit_flip():
    """Test raising an error when converting a Cirq circuit with a bit flip to pyQuil."""

    with pytest.raises(ProgramConversionError):
        q0 = LineQubit(0)
        circuit = Circuit(cirq_ops.bit_flip(p=0.2).on(q0), cirq_ops.measure(q0, key="result"))
        cirq_to_pyquil(circuit)


def test_raise_error_from_pyquil_noisey():
    """Test raising an error when converting a noisey pyQuil program to Cirq."""

    with pytest.raises(ProgramConversionError):
        p = Program()
        p += RX(-np.pi / 2, 0)
        noise_model = _decoherence_noise_model(_get_program_gates(p))
        p = apply_noise_model(p, noise_model)
        pyquil_to_cirq(p)


def test_cirq_to_quil_output_rzz():
    """Test that a RZZ gate is correctly converted to Quil."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[2];
    rzz(pi) q[0], q[1];
    rzz(0) q[0], q[1];
    rzz(pi/2) q[0], q[1];
    """

    p = Program()
    p += CZ(0, 1)
    p += I(0)
    p += I(1)
    p += RZZ(np.pi / 2, 0, 1)
    cirq_circuit = qasm2_to_cirq(qasm)
    p_test: Program = cirq_to_pyquil(cirq_circuit)
    assert p_test.out().replace("pi/2", f"{np.pi/2}") == p.out()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining exceptions for errors raised by qBraid transforms.

"""

from qbraid.exceptions import QbraidError


class CompilationError(QbraidError):
    """Exception raised for errors that occur during the compilation of quantum programs."""


class QasmDecompositionError(QbraidError):
    """Exception raised for errors raised during OpenQASM decomposition processes."""



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for Braket Availability

"""

import datetime
from unittest.mock import Mock, patch

from braket.aws.aws_device import AwsDevice
from braket.device_schema import ExecutionDay

from qbraid.runtime.aws.availability import (
    _calculate_day_factor,
    _current_utc_datetime,
    _is_day_matched,
    _is_time_matched,
    next_available_time,
)


def test_datetime_no_utc_attr():
    """Test _current_utc_datetime when datetime module does not have UTC attribute."""
    with patch("datetime.datetime") as mock_datetime:
        type(mock_datetime).UTC = datetime.timezone.utc
        mock_datetime.utcnow.return_value = "datetime without UTC"
        mock_datetime.now.return_value = "datetime with UTC"
        setattr(datetime, "UTC", datetime.timezone.utc)

        result = _current_utc_datetime()
        assert result == "datetime with UTC"


class ExecutionWindow:
    """Test class for execution window."""

    def __init__(self):
        self.windowStartHour = datetime.time(0)
        self.windowEndHour = datetime.time(23, 59, 59)
        self.executionDay = ExecutionDay.EVERYDAY


class Service:
    """Test class for braket device service."""

    def __init__(self):
        self.executionWindows = [ExecutionWindow()]


class TestProperties:
    """Test class for braket device properties."""

    def __init__(self):
        self.service = Service()


class FakeAwsDevice(AwsDevice):
    """Fake AwsDevice class for testing."""

    def __init__(self, online, available):
        self._is_available = available
        self._status = "ONLINE" if online else "OFFLINE"
        self._properties = TestProperties()

    @property
    def is_available(self):
        """Return availability of the device."""
        return self._is_available

    @is_available.setter
    def is_available(self, value):
        self._is_available = value

    @property
    def status(self):
        """Return status of the device."""
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    @property
    def properties(self):
        """Return properties of the device."""
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value


def test_next_available_time_offline():
    """Test next_available_time when the device is offline."""
    device = FakeAwsDevice(online=False, available=False)
    result = next_available_time(device)
    assert result == (False, "", None)


def test_next_available_time_available():
    """Test next_available_time when the device is available."""
    device = FakeAwsDevice(online=True, available=True)
    result = next_available_time(device)
    assert result == (True, "", None)


def mock_execution_window(execution_day, start_hour, start_minute, end_hour, end_minute):
    """Mock execution window."""
    return Mock(
        executionDay=execution_day,
        windowStartHour=datetime.time(start_hour, start_minute),
        windowEndHour=datetime.time(end_hour, end_minute),
    )


@patch(
    "qbraid.runtime.aws.availability._current_utc_datetime",
    return_value=datetime.datetime(2024, 7, 1, 15, 30, 0, tzinfo=datetime.timezone.utc),
)
def test_next_available_random(mock_utc_datetime):  # pylint: disable=unused-argument
    """Test next_available_time."""
    execution_window = mock_execution_window(ExecutionDay.MONDAY, 16, 0, 17, 0)
    device = Mock(status="ONLINE", is_available=False)
    device.properties.service.executionWindows = [execution_window]

    result = next_available_time(device)
    assert result == (False, "00:30:00", "2024-07-01T16:00:00Z")


def test_next_available_time_no_execution_window():
    """Test next_available_time when there is no execution window."""
    device = Mock(status="ONLINE", is_available=False)
    device.properties.service.executionWindows = []

    result = next_available_time(device)
    assert result == (False, "", None)


ORDERED_DAYS = (
    ExecutionDay.MONDAY,
    ExecutionDay.TUESDAY,
    ExecutionDay.WEDNESDAY,
    ExecutionDay.THURSDAY,
    ExecutionDay.FRIDAY,
    ExecutionDay.SATURDAY,
    ExecutionDay.SUNDAY,
)


def test_is_day_matched():
    """Test _is_day_matched."""
    execution_window = mock_execution_window(ExecutionDay.EVERYDAY, 0, 0, 23, 59)
    for day in range(7):
        assert _is_day_matched(execution_window, day)

    execution_window = mock_execution_window(ExecutionDay.WEEKDAYS, 0, 0, 23, 59)
    for day in range(5):
        assert _is_day_matched(execution_window, day)
    for day in range(5, 7):
        assert not _is_day_matched(execution_window, day)

    execution_window = mock_execution_window(ExecutionDay.WEEKENDS, 0, 0, 23, 59)
    for day in range(5):
        assert not _is_day_matched(execution_window, day)
    for day in range(5, 7):
        assert _is_day_matched(execution_window, day)

    for idx, day in enumerate(ORDERED_DAYS):
        execution_window = mock_execution_window(day, 0, 0, 23, 59)
        for check_day in range(7):
            expected = check_day == idx
            assert (
                _is_day_matched(execution_window, check_day) == expected
            ), f"Failed for {day} with check day {check_day}"


def test_is_time_matched():
    """Test _is_time_matched."""
    execution_window = mock_execution_window(0, 9, 0, 17, 0)
    assert _is_time_matched(execution_window, datetime.time(10, 0))
    assert not _is_time_matched(execution_window, datetime.time(8, 0))
    assert not _is_time_matched(execution_window, datetime.time(18, 0))

    execution_window = mock_execution_window(0, 22, 0, 6, 0)
    assert _is_time_matched(execution_window, datetime.time(23, 0))
    assert _is_time_matched(execution_window, datetime.time(5, 0))
    assert not _is_time_matched(execution_window, datetime.time(7, 0))
    assert not _is_time_matched(execution_window, datetime.time(21, 0))
    assert _is_time_matched(execution_window, datetime.time(22, 0))
    assert _is_time_matched(execution_window, datetime.time(6, 0))
    assert not _is_time_matched(execution_window, datetime.time(6, 1))


@patch(
    "qbraid.runtime.aws.availability._current_utc_datetime",
    return_value=datetime.datetime(2024, 7, 10, 5, 0, 0, tzinfo=datetime.timezone.utc),
)
def test_next_available_time_midnight(mock_utc_datetime):  # pylint: disable=unused-argument
    """Test next_available_time with exec window that spans midnight."""
    execution_window = mock_execution_window(ExecutionDay.WEEKDAYS, 22, 0, 6, 0)
    device = Mock(status="ONLINE", is_available=False)
    device.properties.service.executionWindows = [execution_window]

    result = next_available_time(device)
    assert result == (True, "17:00:00", "2024-07-10T22:00:00Z")


def test_calculate_day_factor():
    """Test _calculate_day_factor."""
    day = 0
    day_factor_0 = _calculate_day_factor(day, datetime.time(0), datetime.time(0))
    assert day_factor_0 == 0

    day = 1
    day_factor_start_before_end = _calculate_day_factor(day, datetime.time(0), datetime.time(1))
    assert day_factor_start_before_end == 0

    day = 2
    day_factor_start_after_end = _calculate_day_factor(day, datetime.time(1), datetime.time(0))
    assert day_factor_start_after_end == 2



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM 2 conversion extras.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

qibo = LazyLoader("qibo", globals(), "qibo")
pyqpanda3 = LazyLoader("pyqpanda3", globals(), "pyqpanda3")

if TYPE_CHECKING:
    import pyqpanda3 as pyqpanda3_  # type: ignore
    import qibo as qibo_  # type: ignore

    from qbraid.programs.typer import Qasm2StringType


@requires_extras("qibo")
def qasm2_to_qibo(qasm: Qasm2StringType) -> qibo_.Circuit:
    """Returns a qibo.Circuit equivalent to the input OpenQASM 2 circuit.

    Args:
        qasm: OpenQASM 2 string to convert to a qibo.Circuit

    Returns:
        qibo.Circuit object equivalent to the input OpenQASM 2 string.
    """
    # Remove problematic comment lines in the qasm code
    lines = [
        line.replace(", ", ",") for line in qasm.split("\n") if not line.strip().startswith("//")
    ]
    # Remove in line problematic comments
    clean_lines = []
    for line in lines:
        clean_line = line.split("//")[0].strip()
        if clean_line:
            clean_lines.append(clean_line)
    qasm = "\n".join(clean_lines)

    return qibo.Circuit.from_qasm(qasm)


@requires_extras("qibo")
def qibo_to_qasm2(circuit: qibo_.Circuit) -> Qasm2StringType:
    """Returns an OpenQASM 2 string equivalent to the input qibo.Circuit.

    Args:
        circuit: qibo.Circuit object to convert to OpenQASM 2 string.

    Returns:
        OpenQASM 2 string equivalent to the input qibo.Circuit.
    """
    return circuit.to_qasm()


@requires_extras("pyqpanda3")
def qasm2_to_pyqpanda3(qasm: Qasm2StringType) -> pyqpanda3_.core.QProg:
    """Returns a pyqpande3.core.QProg equivalent to the input OpenQASM 2 circuit.

    Args:
        qasm: OpenQASM 2 string to convert to a pyqpanda3.core.QProg

    Returns:
        pyqpanda3.core.QProg object equivalent to the input OpenQASM 2 string.
    """
    return pyqpanda3.intermediate_compiler.convert_qasm_string_to_qprog(qasm)


@requires_extras("pyqpanda3")
def pyqpanda3_to_qasm2(circuit: pyqpanda3_.core.QProg) -> Qasm2StringType:
    """Returns an OpenQASM 2 string equivalent to the input pyqpanda3.core.QProg.

    Args:
        circuit: pyqpanda3.core.QProg object to convert to OpenQASM 2 string.

    Returns:
        OpenQASM 2 string equivalent to the input pyqpanda3.core.QProg.
    """
    return pyqpanda3.intermediate_compiler.convert_qprog_to_qasm(circuit)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for calculating Amazon Braket device availability.

"""

import datetime
from typing import Optional

from braket.aws import AwsDevice
from braket.device_schema import DeviceExecutionWindow, ExecutionDay

ORDERED_DAYS = (
    ExecutionDay.MONDAY,
    ExecutionDay.TUESDAY,
    ExecutionDay.WEDNESDAY,
    ExecutionDay.THURSDAY,
    ExecutionDay.FRIDAY,
    ExecutionDay.SATURDAY,
    ExecutionDay.SUNDAY,
)


def _current_utc_datetime() -> datetime.datetime:
    """Return the current UTC datetime."""
    if not hasattr(datetime, "UTC"):  # pragma: no cover
        # backwards compatibility for Python < 3.11
        return datetime.datetime.utcnow()  # pylint: disable=no-member
    return datetime.datetime.now(datetime.timezone.utc)


def _is_day_matched(window: DeviceExecutionWindow, weekday: int):
    """Determine if the current day matches the execution day."""
    if window.executionDay == ExecutionDay.EVERYDAY:
        return True
    if window.executionDay == ExecutionDay.WEEKDAYS:
        return weekday < 5
    if window.executionDay == ExecutionDay.WEEKENDS:
        return weekday > 4
    return (
        window.executionDay in ORDERED_DAYS and ORDERED_DAYS.index(window.executionDay) == weekday
    )


def _is_time_matched(window: DeviceExecutionWindow, current_time_utc: datetime.time):
    """Determine if the current time is within the execution window."""
    if window.windowStartHour < window.windowEndHour:
        return window.windowStartHour <= current_time_utc <= window.windowEndHour
    return current_time_utc >= window.windowStartHour or current_time_utc <= window.windowEndHour


def _calculate_day_factor(day: int, start_time: datetime.time, end_time: datetime.time) -> int:
    """Calculate the day factor based on the current day and execution window."""
    if day == 0:
        return day
    if start_time < end_time:
        return day - 1
    return day


def _calculate_future_time(
    available_time: int, current_datetime_utc: datetime.datetime
) -> tuple[str, str]:
    """
    Calculates future time from the current datetime in UTC and the available time in seconds.

    Args:
        available_time (int): Available time in seconds.
        current_datetime_utc (datetime.datetime): Current datetime in UTC.

    Returns:
        tuple: A tuple containing available time in "HH:MM:SS" format and
               future UTC datetime in ISO 8601 format "YYYY-MM-DDTHH:MM:SSZ".
    """
    hours = available_time // 3600
    minutes = (available_time // 60) % 60
    seconds = available_time % 60
    available_time_hms = f"{hours:02}:{minutes:02}:{seconds:02}"

    time_delta = datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)
    future_time = current_datetime_utc + time_delta
    utc_datetime_str = future_time.strftime("%Y-%m-%dT%H:%M:%SZ")

    return available_time_hms, utc_datetime_str


def next_available_time(device: AwsDevice) -> tuple[bool, str, Optional[str]]:
    """Returns hr/min/sec until device is next available, or empty string if device is offline."""

    is_available_result = False
    available_time = None

    if device.status != "ONLINE":
        return False, "", None

    if device.is_available:
        return True, "", None

    day = 0

    current_datetime_utc = _current_utc_datetime()
    for execution_window in device.properties.service.executionWindows:
        weekday = current_datetime_utc.weekday()
        current_time_utc = current_datetime_utc.time().replace(microsecond=0)

        if (
            execution_window.windowEndHour < execution_window.windowStartHour
            and current_time_utc < execution_window.windowEndHour
        ):
            weekday = (weekday - 1) % 7

        matched_day = _is_day_matched(execution_window, weekday)
        matched_time = _is_time_matched(execution_window, current_time_utc)

        if execution_window.executionDay in ORDERED_DAYS:
            day = (ORDERED_DAYS.index(execution_window.executionDay) - weekday) % 7

        start_time = execution_window.windowStartHour
        end_time = current_time_utc
        day_factor = _calculate_day_factor(day, start_time, end_time)

        td = datetime.timedelta(
            hours=start_time.hour,
            minutes=start_time.minute,
            seconds=start_time.second,
        ) - datetime.timedelta(
            hours=end_time.hour, minutes=end_time.minute, seconds=end_time.second
        )
        days_seconds = 86400 * day_factor
        available_time_secs = td.seconds + days_seconds

        if available_time is None or available_time_secs < available_time:
            available_time = available_time_secs

        is_available_result = is_available_result or (matched_day and matched_time)

    if available_time is None:
        return is_available_result, "", None

    available_time_hms, utc_datetime_str = _calculate_future_time(
        available_time, current_datetime_utc
    )
    return is_available_result, available_time_hms, utc_datetime_str



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining AzureQuantumJob class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from azure.quantum.target.microsoft import MicrosoftEstimatorResult
from azure.quantum.workspace import Workspace

from qbraid._logging import logger
from qbraid.runtime.azure.result_builder import AzureResultBuilder
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.exceptions import JobStateError
from qbraid.runtime.job import QuantumJob
from qbraid.runtime.result import Result
from qbraid.runtime.result_data import AhsResultData, GateModelResultData

from .io_format import OutputDataFormat

if TYPE_CHECKING:
    import azure.quantum


class AzureQuantumJob(QuantumJob):
    """Azure job class."""

    def __init__(self, job_id: str, workspace: Optional[Workspace] = None, **kwargs):
        super().__init__(job_id=job_id, **kwargs)
        self._workspace = workspace or Workspace()
        self._job = self.workspace.get_job(self.id)

    @property
    def workspace(self) -> azure.quantum.Workspace:
        """Return the Azure Quantum Workspace."""
        return self._workspace

    def details(self) -> dict[str, Any]:
        """Return the details of the Azure job and
        update the metadata cache."""
        self._job.refresh()
        details = self._job.details.as_dict()
        self._cache_metadata["details"] = details
        return details

    def status(self) -> JobStatus:
        """Return the current status of the Azure job.

        Returns:
            JobStatus: The current status of the job.
        """
        details: dict = self.details()
        status: str = details.get("status")

        status_map = {
            "Succeeded": JobStatus.COMPLETED,
            "Waiting": JobStatus.QUEUED,
            "Executing": JobStatus.RUNNING,
            "Failed": JobStatus.FAILED,
            "Cancelled": JobStatus.CANCELLED,
            "Finishing": JobStatus.RUNNING,
        }
        return status_map.get(status, JobStatus.UNKNOWN)

    @staticmethod
    def _make_estimator_result(data: dict[str, Any]) -> MicrosoftEstimatorResult:
        """Create a MicrosoftEstimatorResult from the given data.

        Args:
            data (dict): The data to create the result from.

        Returns:
            MicrosoftEstimatorResult: The result created from the data.

        Raises:
            RuntimeError: If the job execution failed.
        """
        if not data["success"]:
            error_data = data["error_data"]
            raise RuntimeError(
                f"Cannot retrieve results as job execution failed "
                f"({error_data['code']}: {error_data['message']})"
            )

        result_data = data["data"]
        return MicrosoftEstimatorResult(result_data)

    def result(self) -> Union[Result, MicrosoftEstimatorResult]:
        """Return the result of the Azure job.

        Returns:
            Union[Result, MicrosoftEstimatorResult]: The result of the job.
        """
        if not self.is_terminal_state():
            logger.info("Result will be available when job has reached final state.")

        job: azure.quantum.Job = self._job

        job.wait_until_completed()

        success = job.details.status == "Succeeded"
        details = job.details.as_dict()

        if job.details.output_data_format == OutputDataFormat.RESOURCE_ESTIMATOR.value:
            return self._make_estimator_result(
                {
                    "job_id": job.id,
                    "target": job.details.target,
                    "job_name": job.details.name,
                    "success": success,
                    "data": job.get_results(),
                    "error_data": (
                        None if job.details.error_data is None else job.details.error_data.as_dict()
                    ),
                }
            )
        builder = AzureResultBuilder(job)
        if job.details.output_data_format == OutputDataFormat.PASQAL.value:
            data = AhsResultData(measurement_counts=builder.get_counts())
        else:
            data = GateModelResultData(measurement_counts=builder.get_counts())
        return Result(
            device_id=job.details.target, job_id=job.id, success=success, data=data, **details
        )

    def cancel(self) -> None:
        """Cancel the Azure job."""
        if self.is_terminal_state():
            raise JobStateError("Cannot cancel; job in terminal state.")

        self._job = self.workspace.cancel_job(self._job)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining CirqCircuit Class

"""

import cirq
import numpy as np

from qbraid.programs.exceptions import ProgramTypeError

from ._model import GateModelProgram


class CirqCircuit(GateModelProgram):
    """Wrapper class for ``cirq.Circuit`` objects."""

    def __init__(self, program: "cirq.Circuit"):
        super().__init__(program)
        if not isinstance(program, cirq.Circuit):
            raise ProgramTypeError(
                message=f"Expected 'cirq.Circuit' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[cirq.Qid]:
        """Return the qubits acted upon by the operations in this circuit"""
        return list(self.program.all_qubits())

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return len(cirq.Circuit(self.program.all_operations()))

    def _unitary(self) -> np.ndarray:
        """Calculate unitary of circuit."""
        return self.program.unitary()

    @staticmethod
    def is_measurement_gate(op: cirq.Operation) -> bool:
        """Returns whether Cirq gate/operation is MeasurementGate."""
        return isinstance(op, cirq.ops.MeasurementGate) or isinstance(
            getattr(op, "gate", None), cirq.ops.MeasurementGate
        )

    @staticmethod
    def _key_from_qubit(qubit: cirq.Qid) -> str:
        if isinstance(qubit, cirq.LineQubit):
            key = str(qubit)
        elif isinstance(qubit, cirq.GridQubit):
            key = str(qubit.row)
        elif isinstance(qubit, cirq.NamedQubit):
            # Only correct if numbered sequentially
            key = str(qubit.name)
        else:
            raise ValueError(
                "Expected qubit of type 'GridQubit' 'LineQubit' or 'NamedQubit'"
                f"but instead got {type(qubit)}"
            )
        return key

    @staticmethod
    def _int_from_qubit(qubit: cirq.Qid) -> int:
        """Extracts an integer value from a given qubit.

        This function interprets the qubit based on its type and extracts an
        integer value accordingly. For a LineQubit, it returns its index.
        For a GridQubit, it returns the row number. For a NamedQubit, it attempts
        to parse and return an integer value from its name. This parsing is only
        successful if the name of the NamedQubit contains an integer. If the name
        does not contain an integer, such as in the case of NamedQubit("q"), or if
        the qubit is not of the supported types (LineQubit, GridQubit, or NamedQubit),
        a ValueError is raised.

        Args:
            qubit (cirq.Qid): The qubit from which to extract an integer.

        Returns:
            int: The extracted integer value from the qubit.

        Raises:
            ValueError: If the qubit is a NamedQubit whose name does not contain
                        an integer, or if the qubit is not of a supported
                        type (LineQubit, GridQubit, or NamedQubit).
        """
        if isinstance(qubit, cirq.LineQubit):
            return int(qubit)
        if isinstance(qubit, cirq.GridQubit):
            return qubit.row
        if isinstance(qubit, cirq.NamedQubit):
            try:
                # NamedQubit("q") will fail here
                return int(qubit._comparison_key().split(":")[0][7:])
            except ValueError as err:
                raise ValueError(
                    f"The provided NamedQubit '{qubit.name}' cannot be converted to an integer."
                ) from err
        else:
            raise ValueError(
                "Expected qubit of type 'GridQubit' 'LineQubit' or 'NamedQubit'"
                f"but instead got {type(qubit)}"
            )

    @staticmethod
    def _make_qubits(qubits: list[cirq.Qid], targets: list[int]) -> list[cirq.Qid]:
        if len(set(type(qubit) for qubit in qubits)) > 1:
            # If mixed types, default to LineQubits
            qubit_type = cirq.LineQubit
        else:
            qubit_type = type(qubits[0])

        if qubit_type == cirq.LineQubit:
            return [cirq.LineQubit(i) for i in targets]
        if qubit_type == cirq.GridQubit:
            rows = [qubit.row for qubit in qubits]
            cols = [qubit.col for qubit in qubits]

            if len(set(rows)) == 1:
                # All qubits in a single row, vary column
                return [cirq.GridQubit(rows[0], i) for i in targets]
            if len(set(cols)) == 1:
                # All qubits in a single column, vary row
                return [cirq.GridQubit(i, cols[0]) for i in targets]

            raise ValueError("GridQubits must be aligned in a single row or column.")
        if qubit_type == cirq.NamedQubit:
            return [cirq.NamedQubit(str(i)) for i in targets]

        raise ValueError(
            f"Expected qubits of type 'GridQubit', 'LineQubit', 'NamedQubit' but got {qubit_type}"
        )

    def _convert_to_line_qubits(self) -> None:
        """Converts a Cirq circuit constructed using NamedQubits to
        a Cirq circuit constructed using LineQubits."""
        qubits = sorted(self.qubits)
        num_grid_qubits = sum(isinstance(qubit, cirq.GridQubit) for qubit in qubits)

        # Check if all qubits are GridQubits
        if len(qubits) == num_grid_qubits:
            rows, cols = zip(
                *((qubit.row, qubit.col) for qubit in qubits if isinstance(qubit, cirq.GridQubit))
            )

            rows_all_equal = len(set(rows)) == 1
            cols_all_equal = len(set(cols)) == 1

            if rows_all_equal:
                qubit_map = {qubit: cirq.LineQubit(qubit.col) for qubit in qubits}
            elif cols_all_equal:
                qubit_map = {qubit: cirq.LineQubit(qubit.row) for qubit in qubits}
            else:
                qubit_map = {qubit: cirq.LineQubit(i) for i, qubit in enumerate(qubits)}
        else:
            qubit_map = {}
            for i, qubit in enumerate(qubits):
                try:
                    qubit_map[qubit] = cirq.LineQubit(self._int_from_qubit(qubit))
                except ValueError:
                    # If _int_from_qubit fails, fallback to default contiguous qubit mapping
                    qubit_map = {qubit: cirq.LineQubit(i) for i, qubit in enumerate(qubits)}
                    break

        self._program = self.program.transform_qubits(lambda q: qubit_map[q])

    def populate_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, adds identity gates to vacant registers as needed."""
        circuit = self.program.copy()
        cirq_qubits = list(circuit.all_qubits())
        if any(isinstance(qubit, cirq.NamedQubit) for qubit in cirq_qubits):
            self._convert_to_line_qubits()
            self.populate_idle_qubits()
        else:
            nqubits = 0
            max_qubit = 0
            occupied_qubits = []
            for qubit in circuit.all_qubits():
                index = self._int_from_qubit(qubit)
                occupied_qubits.append(index)
                max_qubit = max(max_qubit, index)
                nqubits += 1
            qubit_count = max_qubit + 1
            if qubit_count > nqubits:
                all_qubits = list(range(0, qubit_count))
                vacant_qubits = list(set(all_qubits) - set(occupied_qubits))
                cirq_qubits = self._make_qubits(cirq_qubits, vacant_qubits)
                for qubit in cirq_qubits:
                    circuit.append(cirq.I(qubit))
            self._program = circuit

    def remove_idle_qubits(self) -> None:
        """If circuit does not use contiguous qubits/indices, reduces dimension accordingly."""
        qubits = sorted(self.qubits)
        qubit_map = dict(zip(qubits, self._make_qubits(qubits, range(len(qubits)))))
        self._program = self.program.transform_qubits(lambda q: qubit_map[q])

    def reverse_qubit_order(self) -> None:
        """Rerverse qubit ordering of circuit."""
        qubits = self.qubits
        qubits.sort()
        qubits_rev = qubits.copy()
        qubits_rev.reverse()
        qubit_map = dict(zip(qubits, qubits_rev))
        self._program = self.program.transform_qubits(lambda q: qubit_map[q])

    @staticmethod
    def remove_measurements(circuit: cirq.Circuit) -> cirq.Circuit:
        """Remove all measurement gates from the given Cirq circuit.

        Args:
            circuit (cirq.Circuit): The input circuit from which to remove measurement gates.

        Returns:
            cirq.Circuit: A new circuit with all measurement gates removed.
        """
        new_circuit = cirq.Circuit()
        for mom in circuit:
            filtered_operations = [
                op for op in mom.operations if not isinstance(op.gate, cirq.MeasurementGate)
            ]
            if filtered_operations:
                new_circuit.append(cirq.Moment(filtered_operations))
        return new_circuit

    @staticmethod
    def align_final_measurements(circuit: cirq.Circuit) -> cirq.Circuit:
        """
        Align the final measurement gates of all qubits in the same moment
        if they all end with a measurement.

        Args:
            circuit (cirq.Circuit): The input quantum circuit.

        Returns:
            cirq.Circuit: New circuit where all final measurements are aligned in the same moment.
        """
        new_circuit = cirq.Circuit()

        last_ops = {qubit: None for qubit in circuit.all_qubits()}

        for moment in circuit:
            for op in moment.operations:
                if isinstance(op.gate, cirq.MeasurementGate):
                    last_ops[op.qubits[0]] = op

        if all(op is not None for op in last_ops.values()):
            measurement_moment = cirq.Moment()
            for op in last_ops.values():
                measurement_moment = measurement_moment.with_operation(op)
            for moment in circuit:
                non_measurement_moment = cirq.Moment()
                for op in moment.operations:
                    if not isinstance(op.gate, cirq.MeasurementGate):
                        non_measurement_moment = non_measurement_moment.with_operation(op)
                if non_measurement_moment.operations:
                    new_circuit.append(non_measurement_moment)
            new_circuit.append(measurement_moment)
        else:
            return circuit

        return new_circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining AnnealingProblem Class

"""
from __future__ import annotations

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any

from qbraid.programs.program import QuantumProgram
from qbraid.programs.typer import QuboCoefficientsDict

if TYPE_CHECKING:
    import qbraid.runtime


class ProblemType(Enum):
    """Enumeration for different types of annealing models.

    Attributes:
        QUBO: Quadratic Unconstrained Binary Optimization model with binary variables.
        ISING: Ising model with spin variables (-1 or +1).
    """

    QUBO = "qubo"
    ISING = "ising"


@dataclass
class Problem:
    """Represents an annealing problem, including linear and quadratic terms.

    Attributes:
        problem_type: An instance of ProblemType indicating whether the model is QUBO or ISING.
        linear: A dictionary representing the linear coefficients.
        quadratic: A dictionary representing the quadratic coefficients.

    """

    problem_type: ProblemType
    linear: dict[str, float] = field(default_factory=dict)
    quadratic: dict[tuple[str, str], float] = field(default_factory=dict)

    def num_variables(self) -> int:
        """Return the number of variables in the problem."""
        variables = set(self.linear.keys())
        for key in self.quadratic:
            variables.update(key)
        return len(variables)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Problem):
            return False
        if self.problem_type != other.problem_type:
            return False
        if self.linear != other.linear or len(self.quadratic) != len(other.quadratic):
            return False
        for key, value in self.quadratic.items():
            if key in other.quadratic:
                if other.quadratic[key] != value:
                    return False
            elif (key[1], key[0]) in other.quadratic:
                if other.quadratic[(key[1], key[0])] != value:
                    return False
            else:
                return False
        return True


@dataclass
class QuboProblem(Problem):
    """Represents a QUBO problem, subclass of Problem that only includes quadratic coefficients."""

    def __init__(self, coefficients: QuboCoefficientsDict):
        super().__init__(
            problem_type=ProblemType.QUBO,
            linear={},
            quadratic=coefficients,
        )


class AnnealingProgram(QuantumProgram, ABC):
    """Abstract class for annealing problems."""

    @property
    def num_qubits(self) -> int:
        """Number of qubits needed by a quantum device to execute this program."""
        return self.to_problem().num_variables()

    def transform(self, device: qbraid.runtime.QuantumDevice) -> None:
        """Transform program according to device target profile."""
        return None

    @abstractmethod
    def to_problem(self) -> Problem:
        """Return a Problem data class representing this annealing problem."""

    def to_json(self) -> str:
        """Serialize the annealing problem to a JSON string."""
        return json.dumps(self, cls=ProblemEncoder)

    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
        return {"problem": self.to_json()}

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.to_problem() == other.to_problem()


class ProblemEncoder(json.JSONEncoder):
    """Custom JSON encoder for Problem data class."""

    def default(self, o: Any) -> Any:
        if isinstance(o, AnnealingProgram):
            return self.default(o.to_problem())
        if isinstance(o, Problem):
            data = {"problem_type": o.problem_type.value}
            if o.linear:
                data["linear"] = o.linear
            if o.quadratic:
                quadratic_json = {json.dumps(key): value for key, value in o.quadratic.items()}
                data["quadratic"] = quadratic_json
            return data
        if isinstance(o, ProblemType):
            return o.value

        return super().default(o)  # pragma: no cover



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for qbraid.programs.ionq.IonQProgram

"""
import json

import numpy as np
import pytest

from qbraid.interface.random import random_circuit
from qbraid.interface.random.ionq_random import create_gateset_ionq
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.gate_model.ionq import GateSet, InputFormat, IonQProgram
from qbraid.programs.typer import IonQDict


@pytest.fixture
def ionq_dict() -> IonQDict:
    """Return a valid IonQDict object."""
    return {
        "qubits": 3,
        "circuit": [
            {"gate": "h", "target": 0},
            {"gate": "cnot", "control": 0, "target": 1},
            {"gate": "cnot", "control": 0, "target": 2},
        ],
    }


@pytest.fixture
def ionq_program(ionq_dict: IonQDict) -> IonQProgram:
    """Return an IonQProgram object."""
    return IonQProgram(ionq_dict)


def test_ionq_program_bits(ionq_program: IonQProgram):
    """Test the qubits and clbits properties."""
    assert ionq_program.num_qubits == 3
    assert ionq_program.num_clbits == 0


def test_ionq_program_serialize(ionq_program: IonQProgram, ionq_dict: IonQDict):
    """Test the qubits and clbits properties."""
    assert ionq_program.serialize() == {"ionqCircuit": json.dumps(ionq_dict)}


def test_ionq_program_type_error():
    """Test that an error is raised when the input is not an IonQDict object."""
    invalid_program = {"qubits": 3, "circuit": 42}
    with pytest.raises(ProgramTypeError):
        IonQProgram(invalid_program)


def test_ionq_program_determine_gateset_raises():
    """Test that an error is raised when determining
    the gateset of a circuit that mixes native and
    abstract (qis) gates."""
    circuit = [
        {"gate": "h", "target": 0},
        {"gate": "cnot", "control": 0, "target": 1},
        {"gate": "cnot", "control": 0, "target": 2},
        {"gate": "gpi", "phase": 0, "target": 0},
    ]

    with pytest.raises(ValueError) as excinfo:
        IonQProgram.determine_gateset(circuit)
    assert "Invalid gate" in str(excinfo.value)


def test_ionq_program_determine_gateset_native():
    """Test that the native gateset is determined correctly."""
    circuit = [
        {"gate": "gpi", "phase": 0, "target": 0},
        {"gate": "gpi2", "phase": 0, "target": 1},
    ]
    assert IonQProgram.determine_gateset(circuit) == GateSet.NATIVE


def test_ionq_program_determine_gateset_empty_circuit():
    """Test that ValueError is raised when determining the gateset of an empty circuit."""
    with pytest.raises(ValueError) as excinfo:
        _ = IonQProgram.determine_gateset([])
    assert "Circuit is empty. Must contain at least one gate." in str(excinfo.value)


def test_ionq_dict_checks_invalid_gateset_type(ionq_dict):
    """Test that an dict object with an invalid gateset type
    is not considered a valid IonQDict instance."""
    ionq_dict_invalid = ionq_dict.copy()
    ionq_dict_invalid["gateset"] = 42
    assert isinstance(ionq_dict, IonQDict)
    assert not isinstance(ionq_dict_invalid, IonQDict)


def test_ionq_dict_checks_invalid_format(ionq_dict):
    """Test that an dict object with an invalid format type
    is not considered a valid IonQDict instance."""
    ionq_dict_invalid = ionq_dict.copy()
    ionq_dict_invalid["format"] = 42
    assert isinstance(ionq_dict, IonQDict)
    assert not isinstance(ionq_dict_invalid, IonQDict)


def test_input_format_enum():
    """Test the InputFormat enumeration."""
    assert len(list(InputFormat)) == 4
    assert InputFormat.CIRCUIT.value == "ionq.circuit.v0"
    assert InputFormat.QASM.value == "qasm"
    assert InputFormat.OPENQASM.value == "openqasm"
    assert InputFormat.QUIPPER.value == "quipper"


def test_validate_for_gateset(ionq_program: IonQProgram):
    """Test that validate_for_gateset does not raise an error."""
    ionq_program.validate_for_gateset()


def test_validate_for_gateset_raises(ionq_dict: IonQDict):
    """Test that validate_for_gateset raises an error when the circuit contains an invalid gate."""
    ionq_dict_invalid = ionq_dict.copy()
    ionq_dict_invalid["circuit"][0]["gate"] = "not_a_gate"

    ionq_program = IonQProgram(ionq_dict_invalid)

    with pytest.raises(ValueError) as excinfo:
        ionq_program.validate_for_gateset()
    assert "Invalid gate" in str(excinfo.value)


def test_random_circuit_ionq():
    """Test that random_circuit generates a valid IonQ program."""
    rand_program = random_circuit("ionq", max_attempts=1, num_qubits=5, depth=10)
    ionq_program = IonQProgram(rand_program)
    ionq_program.validate_for_gateset()


def test_create_gateset_ionq_max_operands_3_or_more():
    """Test that create_gateset_ionq generates a valid gateset."""
    result = create_gateset_ionq(3)

    assert isinstance(result, np.ndarray)

    assert result.dtype == [("gate", object), ("num_qubits", np.int64), ("num_params", np.int64)]

    expected_gates = {
        "x",
        "y",
        "z",
        "h",
        "s",
        "t",
        "v",
        "si",
        "ti",
        "vi",
        "rx",
        "ry",
        "rz",
        "cx",
        "cy",
        "cz",
        "ch",
        "crx",
        "cry",
        "crz",
        "swap",
        "ccnot",
    }
    actual_gates = set(result["gate"])
    assert actual_gates == expected_gates

    assert len(result) == 22

    assert "ccnot" in result["gate"]

    assert result[result["gate"] == "x"]["num_qubits"][0] == 1
    assert result[result["gate"] == "x"]["num_params"][0] == 0
    assert result[result["gate"] == "rx"]["num_qubits"][0] == 1
    assert result[result["gate"] == "rx"]["num_params"][0] == 1
    assert result[result["gate"] == "cx"]["num_qubits"][0] == 2
    assert result[result["gate"] == "cx"]["num_params"][0] == 0
    assert result[result["gate"] == "crx"]["num_qubits"][0] == 2
    assert result[result["gate"] == "crx"]["num_params"][0] == 1
    assert result[result["gate"] == "ccnot"]["num_qubits"][0] == 3
    assert result[result["gate"] == "ccnot"]["num_params"][0] == 0

    result_4 = create_gateset_ionq(4)
    assert np.array_equal(result, result_4)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QuantumProgram Class

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional

from qbraid._logging import logger

from .alias_manager import get_program_type_alias
from .exceptions import ProgramTypeError
from .experiment import ExperimentType
from .registry import QPROGRAM_REGISTRY
from .spec import ProgramSpec

if TYPE_CHECKING:
    import qbraid.programs
    import qbraid.runtime


class QuantumProgram(ABC):
    """Abstract class for qbraid program wrapper objects."""

    def __init__(self, program: qbraid.programs.QPROGRAM):
        self.spec = self.get_spec(program)
        self._program: Any = None
        self.program = program

    @property
    def program(self) -> qbraid.programs.QPROGRAM:
        """Return the quantum program."""
        return self._program

    @program.setter
    def program(self, value: qbraid.programs.QPROGRAM) -> None:
        """Set the quantum program."""
        expected_type = QPROGRAM_REGISTRY.get(self.spec.alias)
        if not isinstance(value, expected_type):
            raise ProgramTypeError(
                message=(
                    f"Expected program type of '{expected_type}' for "
                    f"derived type alias '{self.spec.alias}'."
                )
            )
        self._program = value

    @property
    def experiment_type(self) -> Optional[ExperimentType]:
        """Returns the ExperimentType corresponding to the sub-module of the program."""
        return self.spec.experiment_type

    @staticmethod
    def get_spec(program: Any) -> ProgramSpec:
        """Return the program spec."""
        try:
            alias = get_program_type_alias(program)
        except ProgramTypeError as err:
            logger.info(err)
            alias = None

        return ProgramSpec(type(program), alias)

    @property
    @abstractmethod
    def num_qubits(self) -> int:
        """Total count of qubits needed by a quantum device
        to execute this program."""

    @abstractmethod
    def transform(self, device: qbraid.runtime.QuantumDevice) -> None:
        """Transform program to according to device target profile."""

    @abstractmethod
    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
